"""
Entrypoint for Gradio, see https://gradio.app/
"""
import sys
import os
from pathlib import Path

# Check if the package is installed or running from source
if __package__ is None:
    # Running from source, add project root to PYTHONPATH
    project_root = str(Path(__file__).parent)
    if project_root not in sys.path:
        sys.path.append(project_root)

from importlib import resources
import platform
import asyncio
import base64
import os
import io
import json
from datetime import datetime
from enum import StrEnum
from functools import partial
from pathlib import Path
from typing import cast, Dict
from PIL import Image

import gradio as gr
from anthropic import APIResponse
from anthropic.types import TextBlock
from anthropic.types.beta import BetaMessage, BetaTextBlock, BetaToolUseBlock
from anthropic.types.tool_use_block import ToolUseBlock

from screeninfo import get_monitors

# TODO: I don't know why If don't get monitors here, the screen resolution will be wrong for secondary screen. Seems there are some conflict with computer_use_ootb_internal.computer_use_demo.tools
screens = get_monitors()
print(screens)
from computer_use_ootb_internal.computer_use_demo.loop import (
    PROVIDER_TO_DEFAULT_MODEL_NAME,
    APIProvider,
    sampling_loop_sync,
)

from computer_use_ootb_internal.computer_use_demo.tools import ToolResult
from computer_use_ootb_internal.computer_use_demo.tools.computer import get_screen_details
SCREEN_NAMES, SELECTED_SCREEN_INDEX = get_screen_details()
# SELECTED_SCREEN_INDEX = None
# SCREEN_NAMES = None

CONFIG_DIR = Path("~/.anthropic").expanduser()
API_KEY_FILE = CONFIG_DIR / "api_key"

WARNING_TEXT = "⚠️ Security Alert: Never provide access to sensitive accounts or data, as malicious web content can hijack Claude's behavior"


class Sender(StrEnum):
    USER = "user"
    BOT = "assistant"
    TOOL = "tool"


def setup_state(state):

    if "messages" not in state:
        state["messages"] = []
    if "model" not in state:
        state["model"] = "gpt-4o + ShowUI"
        # _reset_model(state)
    if "provider" not in state:
        if state["model"] == "qwen2-vl-max + ShowUI":
            state["provider"] = "DashScopeAPI"
        elif state["model"] == "gpt-4o + ShowUI":
            state["provider"] = "openai"
        elif state["model"] == "qwen-vl-7b-instruct + ShowUI":
            state["provider"] = "local-run (not applicable)"
        else:
            state["provider"] = os.getenv("API_PROVIDER", "anthropic") or "anthropic"

    if "provider_radio" not in state:
        state["provider_radio"] = state["provider"]
    
    if "openai_api_key" not in state:  # Fetch API keys from environment variables
        state["openai_api_key"] = os.getenv("OPENAI_API_KEY", "")
    if "anthropic_api_key" not in state:
        state["anthropic_api_key"] = os.getenv("ANTHROPIC_API_KEY", "")    
    if "qwen_api_key" not in state:
        state["qwen_api_key"] = os.getenv("QWEN_API_KEY", "")
    
    # Set the initial api_key based on the provider
    if "api_key" not in state:
        if state["provider"] == "openai":
            state["api_key"] = state["openai_api_key"]
        elif state["provider"] == "anthropic":
            state["api_key"] = state["anthropic_api_key"]
        elif state["provider"] == "qwen":
            state["api_key"] = state["qwen_api_key"]
        else:
            state["api_key"] = ""
    # print(f"state['api_key']: {state['api_key']}")
    if not state["api_key"]:
        print("API key not found. Please set it in the environment or paste in textbox.")

    if "selected_screen" not in state:
        state['selected_screen'] = SELECTED_SCREEN_INDEX if SCREEN_NAMES else 0

    if "auth_validated" not in state:
        state["auth_validated"] = False
    if "responses" not in state:
        state["responses"] = {}
    if "tools" not in state:
        state["tools"] = {}
    if "only_n_most_recent_images" not in state:
        state["only_n_most_recent_images"] = 10 # 10
    if "custom_system_prompt" not in state:
        state["custom_system_prompt"] = load_from_storage("system_prompt") or ""
        # remove if want to use default system prompt
        device_os_name = "Windows" if platform.system() == "Windows" else "Mac" if platform.system() == "Darwin" else "Linux"
        state["custom_system_prompt"] += f"\n\nNOTE: you are operating a {device_os_name} machine"
    if "hide_images" not in state:
        state["hide_images"] = False
    if 'chatbot_messages' not in state:
        state['chatbot_messages'] = []
    


def _reset_model(state):
    state["model"] = PROVIDER_TO_DEFAULT_MODEL_NAME[cast(APIProvider, state["provider"])]


async def main(state):
    """Render loop for Gradio"""
    setup_state(state)
    return "Setup completed"


def validate_auth(provider: APIProvider, api_key: str | None):
    if provider == APIProvider.ANTHROPIC:
        if not api_key:
            return "Enter your Anthropic API key to continue."
    if provider == APIProvider.BEDROCK:
        import boto3

        if not boto3.Session().get_credentials():
            return "You must have AWS credentials set up to use the Bedrock API."
    if provider == APIProvider.VERTEX:
        import google.auth
        from google.auth.exceptions import DefaultCredentialsError

        if not os.environ.get("CLOUD_ML_REGION"):
            return "Set the CLOUD_ML_REGION environment variable to use the Vertex API."
        try:
            google.auth.default(scopes=["https://www.googleapis.com/auth/cloud-platform"])
        except DefaultCredentialsError:
            return "Your google cloud credentials are not set up correctly."


def load_from_storage(filename: str) -> str | None:
    """Load data from a file in the storage directory."""
    try:
        file_path = CONFIG_DIR / filename
        if file_path.exists():
            data = file_path.read_text().strip()
            if data:
                return data
    except Exception as e:
        print(f"Debug: Error loading {filename}: {e}")
    return None


def save_to_storage(filename: str, data: str) -> None:
    """Save data to a file in the storage directory."""
    try:
        CONFIG_DIR.mkdir(parents=True, exist_ok=True)
        file_path = CONFIG_DIR / filename
        file_path.write_text(data)
        # Ensure only user can read/write the file
        file_path.chmod(0o600)
    except Exception as e:
        print(f"Debug: Error saving {filename}: {e}")


def _api_response_callback(response: APIResponse[BetaMessage], response_state: dict):
    response_id = datetime.now().isoformat()
    response_state[response_id] = response


def _tool_output_callback(tool_output: ToolResult, tool_id: str, tool_state: dict):
    tool_state[tool_id] = tool_output


def chatbot_output_callback(message, chatbot_state, hide_images=False, sender="bot"):
    
    def _render_message(message: str | BetaTextBlock | BetaToolUseBlock | ToolResult, hide_images=False):
    
        print(f"_render_message: {str(message)[:100]}")
        
        if isinstance(message, str):
            return message
        
        is_tool_result = not isinstance(message, str) and (
            isinstance(message, ToolResult)
            or message.__class__.__name__ == "ToolResult"
            or message.__class__.__name__ == "CLIResult"
        )
        if not message or (
            is_tool_result
            and hide_images
            and not hasattr(message, "error")
            and not hasattr(message, "output")
        ):  # return None if hide_images is True
            return
        # render tool result
        if is_tool_result:
            message = cast(ToolResult, message)
            if message.output:
                return message.output
            if message.error:
                return f"Error: {message.error}"
            if message.base64_image and not hide_images:
                # somehow can't display via gr.Image
                # image_data = base64.b64decode(message.base64_image)
                # return gr.Image(value=Image.open(io.BytesIO(image_data)))
                return f'<img src="data:image/png;base64,{message.base64_image}">'

        elif isinstance(message, BetaTextBlock) or isinstance(message, TextBlock):
            return message.text
        elif isinstance(message, BetaToolUseBlock) or isinstance(message, ToolUseBlock):
            return f"Tool Use: {message.name}\nInput: {message.input}"
        else:  
            return message

    def _truncate_string(s, max_length=500):
        """Truncate long strings for concise printing."""
        if isinstance(s, str) and len(s) > max_length:
            return s[:max_length] + "..."
        return s
    # processing Anthropic messages
    message = _render_message(message, hide_images)
    
    if sender == "bot":
        chatbot_state.append((None, message))
    else:
        chatbot_state.append((message, None))
    
    # Create a concise version of the chatbot state for printing
    concise_state = [(_truncate_string(user_msg), _truncate_string(bot_msg))
                        for user_msg, bot_msg in chatbot_state]
    # print(f"chatbot_output_callback chatbot_state: {concise_state} (truncated)")

def process_input(user_input, state):
    
    setup_state(state)

    # Append the user message to state["messages"]
    if state["model"] == "gpt-4o + ShowUI" or state["model"] == "qwen2-vl-max + ShowUI" or state["model"] == "qwen-vl-7b-instruct + ShowUI":
        state["messages"].append(
            {
                "role": "user",
                "content": [TextBlock(type="text", text=user_input)],
            }
        )
    elif state["model"] == "claude-3-5-sonnet-20241022":
        state["messages"].append(
            {
                "role": Sender.USER,
                "content": [TextBlock(type="text", text=user_input)],
            }
        )

    # Append the user's message to chatbot_messages with None for the assistant's reply
    state['chatbot_messages'].append((user_input, None))
    yield state['chatbot_messages']  # Yield to update the chatbot UI with the user's message

    # Run sampling_loop_sync with the chatbot_output_callback
    for loop_msg in sampling_loop_sync(
        system_prompt_suffix=state["custom_system_prompt"],
        model=state["model"],
        provider=state["provider"],
        messages=state["messages"],
        output_callback=partial(chatbot_output_callback, chatbot_state=state['chatbot_messages'], hide_images=state["hide_images"]),
        tool_output_callback=partial(_tool_output_callback, tool_state=state["tools"]),
        api_response_callback=partial(_api_response_callback, response_state=state["responses"]),
        api_key=state["api_key"],
        only_n_most_recent_images=state["only_n_most_recent_images"],
        selected_screen=state['selected_screen']
    ):  
        if loop_msg is None:
            yield state['chatbot_messages']
            print("End of task. Close the loop.")
            break
            
        yield state['chatbot_messages']  # Yield the updated chatbot_messages to update the chatbot UI

def launch_app():
    with gr.Blocks(theme=gr.themes.Soft()) as demo:
        state = gr.State({})  # Use Gradio's state management

    setup_state(state.value)  # Initialize the state

    # Retrieve screen details
    gr.Markdown("# Computer Use OOTB")

    if not os.getenv("HIDE_WARNING", False):
        gr.Markdown(WARNING_TEXT)

    with gr.Accordion("Settings", open=True): 
        with gr.Row():
            with gr.Column():
                model = gr.Dropdown(
                    label="Model",
                    choices=["gpt-4o + ShowUI",
                             "claude-3-5-sonnet-20241022",
                             "qwen2-vl-max + ShowUI",
                             "qwen-vl-7b-instruct + ShowUI"],
                    value="gpt-4o + ShowUI",  # Set to one of the choices
                    interactive=True,
                )
            with gr.Column():
                provider = gr.Dropdown(
                    label="API Provider",
                    choices=[option.value for option in APIProvider],
                    value="openai",
                    interactive=False,
                )
            with gr.Column():
                api_key = gr.Textbox(
                    label="API Key",
                    type="password",
                    value=state.value.get("api_key", ""),
                    placeholder="Paste your API key here",
                    interactive=True,
                )
            with gr.Column():
                custom_prompt = gr.Textbox(
                    label="System Prompt Suffix",
                    value="",
                    interactive=True,
                )
            with gr.Column():
                screen_options, primary_index = get_screen_details()
                SCREEN_NAMES = screen_options
                SELECTED_SCREEN_INDEX = primary_index
                screen_selector = gr.Dropdown(
                    label="Select Screen",
                    choices=screen_options,
                    value=screen_options[primary_index] if screen_options else None,
                    interactive=True,
                )
            with gr.Column():
                only_n_images = gr.Slider(
                    label="N most recent screenshots",
                    minimum=0,
                    maximum=10,
                    step=1,
                    value=2,
                    interactive=True,
                )
        # hide_images = gr.Checkbox(label="Hide screenshots", value=False)

    # Define the merged dictionary with task mappings
    with resources.files('computer_use_ootb').joinpath('assets/examples/ootb_examples.json').open('r') as f:
        merged_dict = json.load(f)

    def update_only_n_images(only_n_images_value, state):
        state["only_n_most_recent_images"] = only_n_images_value
    
    # Callback to update the second dropdown based on the first selection
    def update_second_menu(selected_category):
        return gr.update(choices=list(merged_dict.get(selected_category, {}).keys()))

    # Callback to update the third dropdown based on the second selection
    def update_third_menu(selected_category, selected_option):
        return gr.update(choices=list(merged_dict.get(selected_category, {}).get(selected_option, {}).keys()))

    # Callback to update the textbox based on the third selection
    def update_textbox(selected_category, selected_option, selected_task):
        task_data = merged_dict.get(selected_category, {}).get(selected_option, {}).get(selected_task, {})
        prompt = task_data.get("prompt", "")
        preview_image_path = task_data.get("initial_state", "")
        if preview_image_path:
            # Convert Windows path to Unix-style and get just the filename
            filename = os.path.basename(preview_image_path.replace('\\', '/'))
            preview_image = str(resources.files('computer_use_ootb').joinpath('assets/examples/init_states').joinpath(filename))
            print(f"preview_image: {preview_image}")
        else:
            preview_image = None
        task_hint = "Task Hint: " + task_data.get("hint", "")
        return prompt, preview_image, task_hint
    
    # Function to update the global variable when the dropdown changes
    def update_selected_screen(selected_screen_name, state):
        global SCREEN_NAMES
        global SELECTED_SCREEN_INDEX
        SELECTED_SCREEN_INDEX = SCREEN_NAMES.index(selected_screen_name)
        print(f"Selected screen updated to: {SELECTED_SCREEN_INDEX}")
        state['selected_screen'] = SELECTED_SCREEN_INDEX

    def update_model(model_selection, state):
        state["model"] = model_selection
        print(f"Model updated to: {state['model']}")
        
        if model_selection == "qwen2-vl-max + ShowUI":
            provider_choices = ["qwen"]
            provider_value = "qwen"
            provider_interactive = False
            api_key_placeholder = "qwen API key"
        elif model_selection == "gpt-4o + ShowUI":
            # Set provider to "openai", make it unchangeable
            provider_choices = ["openai"]
            provider_value = "openai"
            provider_interactive = False
            api_key_placeholder = "openai API key"
        elif model_selection == "claude-3-5-sonnet-20241022":
            # Provider can be any of the current choices except 'openai'
            provider_choices = [option.value for option in APIProvider if option.value != "openai"]
            provider_value = "anthropic"  # Set default to 'anthropic'
            provider_interactive = True
            api_key_placeholder = "claude API key"
        else:
            # Default case
            provider_choices = [option.value for option in APIProvider]
            provider_value = state.get("provider", provider_choices[0])
            provider_interactive = True
            api_key_placeholder = ""

        # Update the provider in state
        state["provider"] = provider_value
        
        # Update api_key in state based on the provider
        if provider_value == "openai":
            state["api_key"] = state.get("openai_api_key", "")
        elif provider_value == "anthropic":
            state["api_key"] = state.get("anthropic_api_key", "")
        elif provider_value == "qwen":
            state["api_key"] = state.get("qwen_api_key", "")
        else:
            state["api_key"] = ""

        # Use gr.update() instead of gr.Dropdown.update()
        provider_update = gr.update(
            choices=provider_choices,
            value=provider_value,
            interactive=provider_interactive
        )

        # Update the API Key textbox
        api_key_update = gr.update(
            placeholder=api_key_placeholder,
            value=state["api_key"]
        )

        return provider_update, api_key_update
    
    def update_api_key_placeholder(provider_value, model_selection):
        if model_selection == "claude-3-5-sonnet-20241022":
            if provider_value == "anthropic":
                return gr.update(placeholder="anthropic API key")
            elif provider_value == "bedrock":
                return gr.update(placeholder="bedrock API key")
            elif provider_value == "vertex":
                return gr.update(placeholder="vertex API key")
            else:
                return gr.update(placeholder="")
        elif model_selection == "gpt-4o + ShowUI":
            return gr.update(placeholder="openai API key")
        else:
            return gr.update(placeholder="")
        
    def update_system_prompt_suffix(system_prompt_suffix, state):
        state["custom_system_prompt"] = system_prompt_suffix

    with gr.Accordion("Quick Start Prompt", open=False):  # open=False 表示默认收
        # Initialize Gradio interface with the dropdowns
        with gr.Row():
            # Set initial values
            initial_category = "Game Play"
            initial_second_options = list(merged_dict[initial_category].keys())
            initial_third_options = list(merged_dict[initial_category][initial_second_options[0]].keys())
            initial_text_value = merged_dict[initial_category][initial_second_options[0]][initial_third_options[0]]

            with gr.Column(scale=2):
                # First dropdown for Task Category
                first_menu = gr.Dropdown(
                    choices=list(merged_dict.keys()), label="Task Category", interactive=True, value=initial_category
                )

                # Second dropdown for Software
                second_menu = gr.Dropdown(
                    choices=initial_second_options, label="Software", interactive=True, value=initial_second_options[0]
                )

                # Third dropdown for Task
                third_menu = gr.Dropdown(
                    choices=initial_third_options, label="Task", interactive=True, value=initial_third_options[0]
                    # choices=["Please select a task"]+initial_third_options, label="Task", interactive=True, value="Please select a task"
                )

            with gr.Column(scale=1):
                initial_image_value = str(resources.files('computer_use_ootb').joinpath('assets/examples/init_states/honkai_star_rail_showui.png'))
                image_preview = gr.Image(value=initial_image_value, label="Reference Initial State", height=260-(318.75-280))
                hintbox = gr.Markdown("Task Hint: Selected options will appear here.")


        # Textbox for displaying the mapped value
        # textbox = gr.Textbox(value=initial_text_value, label="Action")

    api_key.change(fn=lambda key: save_to_storage(API_KEY_FILE, key), inputs=api_key)

    with gr.Row():
        # submit_button = gr.Button("Submit")  # Add submit button
        with gr.Column(scale=8):
            chat_input = gr.Textbox(show_label=False, placeholder="Type a message to send to Computer Use OOTB...", container=False)
        with gr.Column(scale=1, min_width=50):
            submit_button = gr.Button(value="Send", variant="primary")

    chatbot = gr.Chatbot(label="Chatbot History", autoscroll=True, height=580)
    
    model.change(fn=update_model, inputs=[model, state], outputs=[provider, api_key])
    provider.change(fn=update_api_key_placeholder, inputs=[provider, model], outputs=api_key)
    screen_selector.change(fn=update_selected_screen, inputs=[screen_selector, state], outputs=None)
    only_n_images.change(fn=update_only_n_images, inputs=[only_n_images, state], outputs=None)
    
    # Link callbacks to update dropdowns based on selections
    first_menu.change(fn=update_second_menu, inputs=first_menu, outputs=second_menu)
    second_menu.change(fn=update_third_menu, inputs=[first_menu, second_menu], outputs=third_menu)
    third_menu.change(fn=update_textbox, inputs=[first_menu, second_menu, third_menu], outputs=[chat_input, image_preview, hintbox])

    # chat_input.submit(process_input, [chat_input, state], chatbot)
    submit_button.click(process_input, [chat_input, state], chatbot)

    package_assets_dir = str(resources.files('computer_use_ootb_internal').joinpath('assets'))
    demo.launch(share=True, allowed_paths=[package_assets_dir], server_port=7888)  # TODO: allowed_paths


if __name__ == "__main__":
    launch_app()


import subprocess
import base64
from pathlib import Path
from PIL import ImageGrab
from uuid import uuid4
from screeninfo import get_monitors
import platform
if platform.system() == "Darwin":
    import Quartz  # uncomment this line if you are on macOS
    
from PIL import ImageGrab
from functools import partial
from .base import BaseAnthropicTool, ToolError, ToolResult


OUTPUT_DIR = "./tmp/outputs"

def get_screenshot(selected_screen: int = 0, resize: bool = True, target_width: int = 1920, target_height: int = 1080):
    # print(f"get_screenshot selected_screen: {selected_screen}")
    
    # Get screen width and height using Windows command
    display_num = None
    offset_x = 0
    offset_y = 0
    selected_screen = selected_screen   
    width, height = _get_screen_size()    

    """Take a screenshot of the current screen and return a ToolResult with the base64 encoded image."""
    output_dir = Path(OUTPUT_DIR)
    output_dir.mkdir(parents=True, exist_ok=True)
    path = output_dir / f"screenshot_{uuid4().hex}.png"

    ImageGrab.grab = partial(ImageGrab.grab, all_screens=True)

    # Detect platform
    system = platform.system()

    if system == "Windows":
        # Windows: Use screeninfo to get monitor details
        screens = get_monitors()

        # Sort screens by x position to arrange from left to right
        sorted_screens = sorted(screens, key=lambda s: s.x)

        if selected_screen < 0 or selected_screen >= len(screens):
            raise IndexError("Invalid screen index.")

        screen = sorted_screens[selected_screen]
        bbox = (screen.x, screen.y, screen.x + screen.width, screen.y + screen.height)

    elif system == "Darwin":  # macOS
        # macOS: Use Quartz to get monitor details
        max_displays = 32  # Maximum number of displays to handle
        active_displays = Quartz.CGGetActiveDisplayList(max_displays, None, None)[1]

        # Get the display bounds (resolution) for each active display
        screens = []
        for display_id in active_displays:
            bounds = Quartz.CGDisplayBounds(display_id)
            screens.append({
                'id': display_id,
                'x': int(bounds.origin.x),
                'y': int(bounds.origin.y),
                'width': int(bounds.size.width),
                'height': int(bounds.size.height),
                'is_primary': Quartz.CGDisplayIsMain(display_id)  # Check if this is the primary display
            })

        # Sort screens by x position to arrange from left to right
        sorted_screens = sorted(screens, key=lambda s: s['x'])
        # print(f"Darwin sorted_screens: {sorted_screens}")

        if selected_screen < 0 or selected_screen >= len(screens):
            raise IndexError("Invalid screen index.")

        screen = sorted_screens[selected_screen]
        
        bbox = (screen['x'], screen['y'], screen['x'] + screen['width'], screen['y'] + screen['height'])

    else:  # Linux or other OS
        cmd = "xrandr | grep ' primary' | awk '{print $4}'"
        try:
            output = subprocess.check_output(cmd, shell=True).decode()
            resolution = output.strip().split()[0]
            width, height = map(int, resolution.split('x'))
            bbox = (0, 0, width, height)  # Assuming single primary screen for simplicity
        except subprocess.CalledProcessError:
            raise RuntimeError("Failed to get screen resolution on Linux.")

    # Take screenshot using the bounding box
    screenshot = ImageGrab.grab(bbox=bbox)

    # Set offsets (for potential future use)
    offset_x = screen['x'] if system == "Darwin" else screen.x
    offset_y = screen['y'] if system == "Darwin" else screen.y

    # # Resize if 
    if resize:
        screenshot = screenshot.resize((target_width, target_height))

    # Save the screenshot
    screenshot.save(str(path))

    if path.exists():
        # Return a ToolResult instance instead of a dictionary
        return screenshot, str(path)
    
    raise ToolError(f"Failed to take screenshot: {path} does not exist.")
    
    


def _get_screen_size(selected_screen: int = 0):
    if platform.system() == "Windows":
        # Use screeninfo to get primary monitor on Windows
        screens = get_monitors()

        # Sort screens by x position to arrange from left to right
        sorted_screens = sorted(screens, key=lambda s: s.x)
        if selected_screen is None:
            primary_monitor = next((m for m in get_monitors() if m.is_primary), None)
            return primary_monitor.width, primary_monitor.height
        elif selected_screen < 0 or selected_screen >= len(screens):
            raise IndexError("Invalid screen index.")
        else:
            screen = sorted_screens[selected_screen]
            return screen.width, screen.height
    elif platform.system() == "Darwin":
        # macOS part using Quartz to get screen information
        max_displays = 32  # Maximum number of displays to handle
        active_displays = Quartz.CGGetActiveDisplayList(max_displays, None, None)[1]

        # Get the display bounds (resolution) for each active display
        screens = []
        for display_id in active_displays:
            bounds = Quartz.CGDisplayBounds(display_id)
            screens.append({
                'id': display_id,
                'x': int(bounds.origin.x),
                'y': int(bounds.origin.y),
                'width': int(bounds.size.width),
                'height': int(bounds.size.height),
                'is_primary': Quartz.CGDisplayIsMain(display_id)  # Check if this is the primary display
            })

        # Sort screens by x position to arrange from left to right
        sorted_screens = sorted(screens, key=lambda s: s['x'])

        if selected_screen is None:
            # Find the primary monitor
            primary_monitor = next((screen for screen in screens if screen['is_primary']), None)
            if primary_monitor:
                return primary_monitor['width'], primary_monitor['height']
            else:
                raise RuntimeError("No primary monitor found.")
        elif selected_screen < 0 or selected_screen >= len(screens):
            raise IndexError("Invalid screen index.")
        else:
            # Return the resolution of the selected screen
            screen = sorted_screens[selected_screen]
            return screen['width'], screen['height']

    else:  # Linux or other OS
        cmd = "xrandr | grep ' primary' | awk '{print $4}'"
        try:
            output = subprocess.check_output(cmd, shell=True).decode()
            resolution = output.strip().split()[0]
            width, height = map(int, resolution.split('x'))
            return width, height
        except subprocess.CalledProcessError:
            raise RuntimeError("Failed to get screen resolution on Linux.")



import json
import uiautomation as auto
import time
import subprocess
import pygetwindow as gw
from pywinauto import Application, Desktop
import os
import re
import datetime
import requests
import time
from termcolor import colored
import subprocess
import win32gui
import win32process
import psutil
from PIL import Image, ImageDraw
import base64
import pickle
import requests
from PIL import Image, ImageDraw
# from pywinauto.application import Application
from pywinauto.findwindows import find_windows


software_name_map = {"adobe_acrobat": "Adobe Acrobat", "premiere_pro": "Adobe Premiere Pro"}
    
def get_screenshot(software_name=None):
    software_name = software_name_map.get(software_name, software_name)
    gui = GUICapture()
    meta_data, screenshot_path = gui.capture(software=software_name)
    return meta_data, screenshot_path


def get_control_properties(control, properties_list, no_texts=False):
    prop_dict = {}
    for prop in properties_list:
        # Skip 'texts' property if no_texts flag is set
        if no_texts and prop in ['texts', "friendly_class_name", "automation_id", "rectangle", "UIA_ImageControlTypeId"]:
            prop_dict[prop] = ['']  # Use an empty list as a placeholder
            continue

        # Check if the control has the property as an attribute
        if hasattr(control, prop):
            attr = getattr(control, prop)
            # Ensure the attribute is callable before attempting to call it
            if callable(attr):
                try:
                    value = attr()
                    # Special handling for 'rectangle' property
                    if prop == 'rectangle':
                        value = [value.left, value.top, value.right, value.bottom]
                    prop_dict[prop] = value
                except Exception:
                    # If there's an exception, such as a method failing, we skip it
                    continue
            else:
                # If the attribute is not callable, directly assign it
                prop_dict[prop] = attr
    return prop_dict


class GUICapture:
    """
    A class to capture and interact with a GUI of a specified application.
    """
    BLUE = '\033[94m'
    GREEN = '\033[92m'
    CYAN = '\033[96m'
    RED = '\033[91m'
    END = '\033[0m'

    def __init__(self, cache_folder='.cache/'):
        """
        Initialize the GUICapture instance.
        """
        self.task_id = self.get_current_time()
        self.cache_folder = os.path.join(cache_folder, self.task_id)
        os.makedirs(self.cache_folder, exist_ok=True)
        self.current_step = 0
        self.history = []
        self.port = 6007      

    def capture(self, software=None):
        """
        Execute the capture process.
        """
        start = time.time()
        # time.sleep(2)  # Consider explaining why this delay is necessary

        if software is None:
            software = self.get_topest_window()

        self.connect_to_application(software)
        meta_data = self.get_gui_meta_data(software)
        screenshot_path = self.capture_screenshot()
        print(f"Time used: {time.time() - start}")
        start = time.time()
        return meta_data, screenshot_path
    
    def get_topest_window(self):
        # Find the window of interest (replace 'Window Title' with the actual title)
        all_window_titles = gw.getAllTitles()

        for window_title in all_window_titles:
            # if 'Window Title' in window:
            win = gw.getWindowsWithTitle(window_title)[0]
            if window_title != "":
                print(f"Top position of the window '{window_title}': {win.top}")
                break

        return window_title
    
    def connect_to_application(self, software_name):
        """
        Connect to the target application.
        """
        try:
            window_handles = find_windows(title_re=f".*{software_name}*", visible_only=False)
            # print(window_handles)
            self.app = Application(backend="uia").connect(handle=window_handles[0])
            # self.app = Application(backend="uia").connect(title_re=f".*{software_name}*")
        except Exception as e:
            print(f"Error connecting to application: {e}")
            try:
                print("Try to connect to the application by using the window name.")
                self.app = self.detect_duplicate_name_windows(software_name)
            except Exception as e:
                print(f"Error connecting to application: {e}")

    def detect_duplicate_name_windows(self, software_name):
        # 使用find_windows函数查找所有匹配的窗口句柄
        window_handles = find_windows(title_re=f".*{software_name}*", visible_only=False)

        # 检查找到的窗口句柄数量
        if window_handles:
            # 假设我们想要连接到找到的第一个窗口
            first_window_handle = window_handles[0]

            # 使用窗口句柄连接到应用程序
            app = Application(backend="uia").connect(handle=first_window_handle)
            # app = Application(backend="uia").connect(handle=first_window_handle)

            return app
            # # 通过app对象操作窗口
            # # 例如，获取窗口的主窗口对象并进行操作
            # main_window = app.window(handle=first_window_handle)
            # # 现在可以对main_window进行操作，例如点击按钮、输入文本等

        else:
            print("没有找到匹配的窗口")
            return None

    def get_gui_meta_data(self, software):
        # Connect to the application
        # Initialize data storage
        control_properties_list = ['friendly_class_name', 'texts', 'rectangle', 'automation_id']
        th = 1000 # software_th.get(software, 100)

        def recurse_controls(control, current_depth=0):

            # from IPython.core.debugger import Pdb; Pdb().set_trace()

            children = control.children()

            child_data = []
            if current_depth > th:
                return []
        
            for child in children:
                # check if the control is visible
                if not child.is_visible():
                    continue  

                properties = get_control_properties(child, control_properties_list)

                # Check if the control is a ComboBox, which may encounter bug while acquire text
                no_texts = True if properties.get('friendly_class_name') == 'ComboBox' else False
                # no_texts = False
                    
                child_data.append({
                    'properties': get_control_properties(child, control_properties_list, no_texts=no_texts),
                    'children': recurse_controls(child, current_depth + 1)
                })

            return child_data

        all_windows = self.app.windows()
        window_names = [window.window_text() for window in all_windows]
        meta_data = {}
        for window_name in window_names:
            # from IPython.core.debugger import Pdb; Pdb().set_trace()

            if window_name:
                print(window_name)
                target_window = self.app.window(title=window_name)
                print(target_window)
                # target_window.set_focus()

                # Traverse the control tree
                meta_data[window_name] = recurse_controls(target_window)

        return meta_data

    def capture_screenshot(self, save_path=None):
        # save screenshot and return path
        if save_path:
            screenshot_path = save_path
        else:
            screenshot_path = os.path.join(self.cache_folder, f'screenshot-{self.current_step}.png')

        screenshot = auto.GetRootControl().ToBitmap()
        screenshot.ToFile(screenshot_path)
        return screenshot_path

    @staticmethod
    def get_current_time():
        return datetime.datetime.now().strftime('%Y%m%d_%H%M%S')


def get_all_windows():
    all_windows = gw.getAllWindows()
    all_windows_name = [win.title for win in all_windows if win.title]
    all_windows_name = simplify_window_names(all_windows_name)
    return all_windows_name


def simplify_window_names(names):
    simplified_names = []
    for name in names:
        # Split the name by '-' and strip whitespace
        parts = [part.strip() for part in name.split('-')]
        # Use the part after the last '-' if available, otherwise the original name
        simplified_name = parts[-1] if len(parts) > 1 else name
        simplified_names.append(simplified_name)
    return simplified_names


def maximize_window(title):
    """Maximize the window"""
    windows = gw.getWindowsWithTitle(title)
    if windows:
        window = windows[0]  # Assume the first window is target window
        if not window.isMaximized:
            window.maximize()
            print(f"Window '{title}' has been maximized.")
        else:
            print(f"Window '{title}' is already maximized.")
    else:
        print(f"No window with the title '{title}' found.")


def extract_texts_and_rectangles(data):
    result = []
    
    def traverse(node):
        # Extract the current node's text and rectangle
        if 'properties' in node:
            texts = node['properties'].get('texts', [])
            rectangle = node['properties'].get('rectangle', [])
            if texts and len(texts) > 0 and rectangle:
                result.append([texts[0], rectangle])
        
        # Traverse children if they exist
        if 'children' in node:
            for child in node['children']:
                traverse(child)
    
    traverse(data[0])  # Assuming the data is a list with one top-level item
    return result


def visualize(gui, screenshot_path, if_show=True):
    ui_elements = []

    def extract_elements(node):
        if isinstance(node, list):
            for item in node:
                extract_elements(item)
        elif isinstance(node, dict):
            properties = node.get('properties', {})
            texts = properties.get('texts', [])
            rectangle = properties.get('rectangle', [])
            if rectangle and texts:
                # Use the first text as the name or join all texts
                # name = texts[0] if texts else ''
                name = texts
                ui_elements.append((name, rectangle))
            # Recurse into children
            children = node.get('children', [])
            extract_elements(children)

    for window_name, panels in gui.items():
        extract_elements(panels)

    image = Image.open(screenshot_path)
    draw = ImageDraw.Draw(image)

    # Draw rectangles and text
    for element in ui_elements:
        name, rectangle = element
        try:
            # print(type(name))
            if isinstance(name, list) and len(name) == 1:
                name = name[0].encode('latin-1', 'ignore').decode('latin-1')
            else:
                name = name.encode('latin-1', 'ignore').decode('latin-1')
                # name = name[0].encode('latin-1', 'ignore').decode('latin-1')
        except:
            print(name)
            name = ""
        # Draw rectangle in red
        try:
            draw.rectangle(rectangle, outline="red")
            # Calculate text position above the rectangle
            text_position = (rectangle[0]-15, rectangle[1] - 15)  # Positioned above the rectangle
            draw.text(text_position, name, fill="red")  # Default font
        except:
            print(rectangle)
            print(name)


    # if if_show:
    #     display(image)
    return image


def encode_image(image_path):
    """Encode image file to base64."""
    with open(image_path, "rb") as image_file:
        return base64.b64encode(image_file.read()).decode("utf-8")
    
def encode_task(task):
    if isinstance(task, str):
        return task
    else:
        return base64.b64encode(pickle.dumps(task)).decode('utf-8')
    
def send_gui_parser_request(url, software_name, screenshot_path, meta_data, task_id=None, step_id=None):
    """Send a POST request to the server with the query and the image."""
    # Encode the image
    screenshot_data = encode_image(screenshot_path)
    
    # Construct the request data
    data = {
        "screenshot": screenshot_data,
        "GUI": meta_data,
        "software_name": software_name,
        "task_id": task_id,
        "step_id": step_id
    }
    
    # Send POST request
    response = requests.post(url, json=data)
    return response.json()


def send_actor_request(url, 
                       current_task, 
                       parsed_screenshot,
                       screenshot_path: str, 
                       software_name: str, 
                       history: list=[], 
                       error_message: str="", 
                       next_step_tip: str="", 
                       pre_act_success_flag: bool=False, 
                       pre_act_resume_flag: bool=False, 
                       task_id: str=None, 
                       step_id: str=None):
    """Send a POST request to the server with the query and the image."""
    # Encode the image
    screenshot_data = encode_image(screenshot_path)
        
    # Construct the request data
    data = {
        "current_task": encode_task(current_task),
        "parsed_screenshot": parsed_screenshot,
        "screenshot": screenshot_data,
        "history": history,
        "error_message": error_message,
        "next_step": next_step_tip,
        "pre_act_success_flag": pre_act_success_flag,
        "pre_act_resume_flag": pre_act_resume_flag,
        "software_name": software_name,
        "task_id": task_id,
        "step_id": step_id, 
    }
    
    # Send POST request
    response = requests.post(url, json=data)
    return response.json()

if __name__ == '__main__':
    capture = GUICapture('Adobe Premiere Pro')
    meta_data, screenshot_path = capture.run("None", send_data=False)



import os
import ast
import base64
from io import BytesIO
from pathlib import Path
from uuid import uuid4

import pyautogui
import requests
import torch
from PIL import Image, ImageDraw
from qwen_vl_utils import process_vision_info
from transformers import AutoProcessor, Qwen2VLForConditionalGeneration

from computer_use_ootb_internal.computer_use_demo.gui_agent.llm_utils.oai import encode_image
from computer_use_ootb_internal.computer_use_demo.tools.colorful_text import colorful_text_showui, colorful_text_vlm
from computer_use_ootb_internal.computer_use_demo.tools.screen_capture import get_screenshot

os.environ["TOKENIZERS_PARALLELISM"] = "false"


class ShowUIActor:
    _NAV_SYSTEM = """You are an assistant trained to navigate the {_APP} screen. 
    Given a task instruction, a screen observation, and an action history sequence, 
    output the next action and wait for the next observation. 
    Here is the action space:
    {_ACTION_SPACE}
    """

    _NAV_FORMAT = """
    Format the action as a dictionary with the following keys:
    {'action': 'ACTION_TYPE', 'value': 'element', 'position': [x,y]}
    
    If value or position is not applicable, set it as None.
    Position might be [[x1,y1], [x2,y2]] if the action requires a start and end position.
    Position represents the relative coordinates on the screenshot and should be scaled to a range of 0-1.
    """

    action_map = {
    'web': """
        1. CLICK: Click on an element, value is not applicable and the position [x,y] is required. 
        2. INPUT: Type a string into an element, value is a string to type and the position [x,y] is required. 
        3. SELECT: Select a value for an element, value is not applicable and the position [x,y] is required. 
        4. HOVER: Hover on an element, value is not applicable and the position [x,y] is required.
        5. ANSWER: Answer the question, value is the answer and the position is not applicable.
        6. ENTER: Enter operation, value and position are not applicable.
        7. SCROLL: Scroll the screen, value is the direction to scroll and the position is not applicable.
        8. SELECT_TEXT: Select some text content, value is not applicable and position [[x1,y1], [x2,y2]] is the start and end position of the select operation.
        9. COPY: Copy the text, value is the text to copy and the position is not applicable.
        10. ESC: ESCAPE operation, value and position are not applicable.
        11. PRESS: Long click on an element, value is not applicable and the position [x,y] is required. 
        """,

    'phone': """
        1. INPUT: Type a string into an element, value is not applicable and the position [x,y] is required. 
        2. SWIPE: Swipe the screen, value is not applicable and the position [[x1,y1], [x2,y2]] is the start and end position of the swipe operation.
        3. TAP: Tap on an element, value is not applicable and the position [x,y] is required.
        4. ANSWER: Answer the question, value is the status (e.g., 'task complete') and the position is not applicable.
        5. ENTER: Enter operation, value and position are not applicable.
        """
    }

    def __init__(self, model_path, output_callback, device=torch.device("cpu"), split='web', selected_screen=0):
        self.device = device
        self.split = split
        self.selected_screen = selected_screen
        self.output_callback = output_callback
        
        if not model_path:
            model_path = "showlab/ShowUI-2B"
        
        self.model = Qwen2VLForConditionalGeneration.from_pretrained(
            model_path,
            torch_dtype=torch.bfloat16,
            device_map="cpu"
        ).to(self.device)
        
        self.min_pixels = 256 * 28 * 28
        self.max_pixels = 1344 * 28 * 28
        self.processor = AutoProcessor.from_pretrained(
            "Qwen/Qwen2-VL-2B-Instruct",
            # "./Qwen2-VL-2B-Instruct",
            min_pixels=self.min_pixels,
            max_pixels=self.max_pixels
        )
        self.system_prompt = self._NAV_SYSTEM.format(
            _APP=split,
            _ACTION_SPACE=self.action_map[split]
        )
        self.action_history = ''  # Initialize action history

    def __call__(self, messages):

        task = messages
        
        # screenshot
        screenshot, screenshot_path = get_screenshot(selected_screen=self.selected_screen, resize=True, target_width=1280, target_height=720)
        screenshot_path = str(screenshot_path)
        image_base64 = encode_image(screenshot_path)
        self.output_callback(f'Screenshot for {colorful_text_showui}:\n<img src="data:image/png;base64,{image_base64}">', sender="bot")

        # Use system prompt, task, and action history to build the messages
        messages_for_processor = [
            {
                "role": "user",
                "content": [
                    {"type": "text", "text": self.system_prompt},
                    {"type": "image", "image": screenshot_path, "min_pixels": self.min_pixels, "max_pixels": self.max_pixels},
                    {"type": "text", "text": f"Task: {task}"}
                    # {"type": "text", "text": f"\nAction History: {self.action_history}}
                ],
            }
        ]
        
        text = self.processor.apply_chat_template(
            messages_for_processor, tokenize=False, add_generation_prompt=True,
        )
        image_inputs, video_inputs = process_vision_info(messages_for_processor)
        inputs = self.processor(
            text=[text],
            images=image_inputs,
            videos=video_inputs,
            padding=True,
            return_tensors="pt",
        )
        inputs = inputs.to(self.device)
        
        generated_ids = self.model.generate(**inputs, max_new_tokens=128)
        generated_ids_trimmed = [
            out_ids[len(in_ids):] for in_ids, out_ids in zip(inputs.input_ids, generated_ids)
        ]
        output_text = self.processor.batch_decode(
            generated_ids_trimmed, skip_special_tokens=True, clean_up_tokenization_spaces=False
        )[0]

        # Update action history
        self.action_history += output_text + '\n'

        # Return response in expected format
        response = {'content': output_text, 'role': 'assistant'}
        return response



    def parse_showui_output(self, output_text):
        try:
            # Ensure the output is stripped of any extra spaces
            output_text = output_text.strip()

            # Wrap the input in brackets if it looks like a single dictionary
            if output_text.startswith("{") and output_text.endswith("}"):
                output_text = f"[{output_text}]"

            # Validate if the output resembles a list of dictionaries
            if not (output_text.startswith("[") and output_text.endswith("]")):
                raise ValueError("Output does not look like a valid list or dictionary.")

            # Parse the output using ast.literal_eval
            parsed_output = ast.literal_eval(output_text)

            # Ensure the result is a list
            if isinstance(parsed_output, dict):
                parsed_output = [parsed_output]
            elif not isinstance(parsed_output, list):
                raise ValueError("Parsed output is neither a dictionary nor a list.")

            # Ensure all elements in the list are dictionaries
            if not all(isinstance(item, dict) for item in parsed_output):
                raise ValueError("Not all items in the parsed output are dictionaries.")

            return parsed_output

        except Exception as e:
            print(f"Error parsing output: {e}")
            return None



from .text_detection import text_detection
from .utils import *
from .panel_recognition.llm_panel_recognize import recognize_panel
from PIL import Image
from concurrent.futures import ThreadPoolExecutor
from .icon_detection.icon_detection import IconDetector
import glob


class AdobeParser():
    name = "adobe_parser"

    def __init__(self, cache_folder=".cache/"):
        self.cache_folder = cache_folder
        self.task_id = get_current_time()
        self.action_type = ["click"]

    def __call__(
        self, meta_data, screenshot_path, window_name=None, scaleFactor=None
    ):
        self.window_name = window_name.lower()
        self.scaleFactor = scaleFactor
        ocr = text_detection(screenshot_path, save_png=False)
        highlight_ocr = self.detect_highlight_with_ocr(screenshot_path)

        self.parsed_gui = {window_name: []}
        for window_name, window_meta_data in meta_data.items():
            if not window_meta_data:
                continue

            if window_meta_data[0]["properties"]["friendly_class_name"] in [
                "MenuItem",
                "Edit",
                "Button",
            ]:
                self.parsed_gui[window_name] += self.parse_popup_window(
                    window_meta_data, window_name
                )
            else:
                self.parsed_gui[window_name] += self.parse_main_window(
                    window_meta_data, screenshot_path, ocr, highlight_ocr
                )
                self.parsed_gui[window_name] += self.parse_menu(window_meta_data)

        return self.parsed_gui

    @staticmethod
    def parse_menu(meta_data):
        menu_items = []
        for item in meta_data:
            if item["properties"]["friendly_class_name"] == "Menu":
                new_item = {
                    "name": item["properties"]["texts"][0],
                    "rectangle": item["properties"]["rectangle"],
                    "type": ["click"],
                    "elements": [
                        [
                            {
                                "name": child["properties"]["texts"][0],
                                "rectangle": child["properties"]["rectangle"],
                                "type": ["click"],
                            }
                            for child in item["children"]
                            if child["properties"]["texts"]
                        ]
                    ],
                }
                menu_items.append(new_item)
        return menu_items

    def parse_popup_window(self, meta_data, window_name):
        filtered_data = self.filter_data(meta_data)
        sorted_data_y = self.sort_data_by_y_coordinate(filtered_data)
        elements = self.organize_elements(sorted_data_y)

        for entry in elements:
            for item in entry:
                item["type"] = self.action_type

        panel_item = {
            "name": window_name,
            "rectangle": find_compact_bounding_box(elements),
            "elements": elements,
        }
        return [panel_item]

    def parse_main_window(self, window_meta_data, screenshot_path, ocr, highlight_ocr):
        screen_resolution = Image.open(screenshot_path).size
        params = [
            (ocr, highlight_ocr, raw_item, screenshot_path, screen_resolution)
            for raw_item in window_meta_data
            if raw_item["properties"]["friendly_class_name"] in ["Pane", "Dialog"]
        ]

        main_panel = self.run_parallel_tasks(self.parse_panel, params)
        return main_panel

    def parse_panel(
        self, ocr, highlight_ocr, raw_item, screenshot_path, screen_resolution
    ):
        panel_name = recognize_panel(
            ocr, highlight_ocr, raw_item, screen_resolution, self.window_name
        )
        raw_item["properties"]["rectangle"] = [
            max(0, value) for value in raw_item["properties"]["rectangle"]
        ]

        panel_item = {
            "name": panel_name,
            "rectangle": raw_item["properties"]["rectangle"],
            "type": self.action_type + ["doubleClick", "rightClick"],
        }

        temp = {
            "editing_control": self.get_text(panel_item, ocr, screenshot_path),
            "button": self.get_button(panel_item, screenshot_path, self.scaleFactor)
        }
        
        
        panel_item["elements"] = self.merge_elements(temp)
        return panel_item


    def get_button(self, panel_item, screenshot_path, scale_factor="1.0x"):
        # print("GXW",scale_factor)
        # crop the panel based on the rectangle
        # print("processing button fuck")
        print(panel_item["name"], panel_item["rectangle"])
        panel_img = crop_panel(panel_item["rectangle"], screenshot_path)
        # get the button
        # th = 0.90 if self.software_name == 'after effect' else 0.78

        th = 0.80 if self.software_name == "after effect" else 0.75
        if self.software_name in [
            "premiere",
            "after effect",
            "Premiere",
            "After Effect",
        ]:
            print("Screenshot:", screenshot_path)
            button_box = self.detect_button_adobe(
                panel_img,
                software_name=self.software_name,
                panel_name=panel_item["name"],
                threshold=th,
                scale_factor=scale_factor,
            )
            
        # restore the button coordinate to the whole screenshot
        button_box = restore_coordinate(button_box, panel_item["rectangle"])
        return button_box
    
    @staticmethod
    def load_icon_templates(icon_folder, software_name, panel_name, scale_factor="1.0x"):
        # 初始化空的路径列表
        if "Accessory" in panel_name:
            panel_name = "Accessory"
        # 初始化空的路径列表
        if panel_name:
            template_folder = f"{icon_folder}/{software_name}/{panel_name}"
        else:
            template_folder = f"{icon_folder}/{software_name}"

        print("GXW load_icon_templates:scale_factor", scale_factor)
        template_folder = f"{icon_folder}/{software_name}/{scale_factor}/{panel_name}"
        print("loading icon templates... from ", template_folder)

        icon_path = glob.glob(f"{template_folder}/**/*.png", recursive=True)
        print("found ", len(icon_path), " icons")

        icons = []
        for template_path in icon_path:
            # 读取模板图片
            template = cv2.imread(template_path, cv2.IMREAD_COLOR)
            # 获取模板图片的名称
            # print(template_path)
            name = re.search(r"[^\\/]+(?=\.\w+$)", template_path).group(0)
            name = re.sub(r"^\d+_", "", name) + "_icon"
            # print(name)
            # 将模板图片的名称和图片加入到列表中
            icons.append({"name": name, "template": template, "path": template_path})
        return icons
    
    @staticmethod
    def preprocess_image(img, software_name):
        # 转换为灰度
        # gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        # 二值化
        # _, binary = cv2.threshold(gray, 50, 255, cv2.THRESH_BINARY)
        if software_name in ["premiere", "after effect"]:
            threshold = 60
        elif software_name in ["adobe acrobat", "acrobat"]:
            threshold = 190
        elif software_name in ["word", "excel", "powerpoint"]:
            threshold = 190
        else:
            threshold = 130

        binary, saved_path = multivalue_image(
            img,
            mode="None",  # 或者您可以用其他任何字符串，这里不重要
            thresholds=[threshold],  # 一个单一的阈值
            interval_values=[0, 255],  # 两个区间值
            save=False,  # 是否保存图像
            cache_folder="./.cache",  # 缓存文件夹
        )
        return binary
    @staticmethod
    def divide_activated_area(image):
        hsv = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
        # Define the scope of blue
        lower_blue = np.array([105, 100, 100])
        upper_blue = np.array([130, 255, 255])
        # Creates a mask that sets parts within a specified color range to white and other parts to black
        mask = cv2.inRange(hsv, lower_blue, upper_blue)
        # Perform a bitwise AND operation on the original image and the mask to extract the blue part
        blue = cv2.bitwise_and(image, image, mask=mask)
        non_blue_part = cv2.bitwise_not(image, image, mask)
        return blue, non_blue_part
    
    
    def detect_button_adobe(
        self,
        image,
        software_name="premiere",
        panel_name=None,
        icon_folder="./cache",
        icon_type=None,
        threshold=0.75,
        scale_factor="1x",
    ):
        # image的格式 cv2.imread(image_path, cv2.IMREAD_COLOR)   最理想情况是1080p的图片否则质量不会太好
        # button_folder button 库位置
        # threshold 模版匹配的阈值， 越高越准确，但是button数目有时会变少

        # image_path = "extracted_img/frame9725.jpg"
        # image = cv2.imread(image_path, cv2.IMREAD_COLOR)

        # print("GXW detect_button_adobe:scale_factor",scale_factor)

        blue_image, non_blue_image = self.divide_activated_area(image)
        non_blue_image = self.preprocess_image(non_blue_image, software_name)
        if "Accessory" in software_name:
            panel_name = "Accessory"

        templates = self.load_icon_templates(
            icon_folder, software_name, panel_name, scale_factor
        )
        # print("templates", templates)
        all_boxes, all_scores, labels = [], [], []
        # count = 0
        for i, template in enumerate(templates):
            icon_name = template["name"]
            icon_template = template["template"]

            icon_template_blue, icon_template_non_blue = self.divide_activated_area(
                icon_template
            )
            icon_template_non_blue = self.preprocess_image(icon_template_non_blue, software_name)

            # find the best scale for the template at the first iteration
            if i == 0:
                best_scale_blue = 1
                if "activated" in icon_name:
                    best_scale_blue = self.get_best_matching_scale(
                        blue_image, icon_template_blue
                    )
                best_scale_non_blue = self.get_best_matching_scale(
                    non_blue_image, icon_template_non_blue
                )

            if "activated" in icon_name:
                matches_blue, scores_blue = self.multi_scale_template_matching(
                    blue_image,
                    icon_template_blue,
                    threshold=threshold,
                    scales=[best_scale_blue],
                )

                icon_width = icon_template_blue.shape[1]
                icon_height = icon_template_blue.shape[0]
                for match, score in zip(matches_blue, scores_blue):
                    (pt_x, pt_y), scale = match

                    end_x = int(pt_x + icon_width * scale)
                    end_y = int(pt_y + icon_height * scale)

                    # 保存所有的框到all_boxes
                    all_boxes.append([pt_x, pt_y, end_x, end_y])
                    all_scores.append(score)
                    labels.append(icon_name)

            if "activated" not in icon_name:
                matches_non_blue, scores_non_blue = self.multi_scale_template_matching(
                    non_blue_image,
                    icon_template_non_blue,
                    threshold=threshold,
                    scales=[best_scale_non_blue],
                )
                icon_width = icon_template_non_blue.shape[1]
                icon_height = icon_template_non_blue.shape[0]
                for match, score in zip(matches_non_blue, scores_non_blue):
                    (pt_x, pt_y), scale = match

                    end_x = int(pt_x + icon_width * scale)
                    end_y = int(pt_y + icon_height * scale)

                    # 保存所有的框到all_boxes
                    all_boxes.append([pt_x, pt_y, end_x, end_y])
                    all_scores.append(score)
                    labels.append(icon_name)

        # print(threshold, labels)
        # 应用NMS bbox 去重
        nms_boxes, pick = self.non_max_suppression(all_boxes, 0.15, all_scores)
        labels = [labels[i] for i in pick]

        button_items = []
        for ix, box in enumerate(nms_boxes):
            if "scroll bar" in labels[ix] or "effects submenu" in labels[ix]:
                item = {
                    "name": labels[ix],
                    "rectangle": list(box),
                }
            else:
                item = {
                    "name": labels[ix],
                    "rectangle": list(box),
                }
            button_items.append(item)

        # button_items = [{"name": labels[ix], "rectangle": list(box), 'type': ['moveTo', 'click']} for ix, box in enumerate(nms_boxes)]

        return button_items

    @staticmethod
    def get_best_matching_scale(image, template, threshold=0.8, scales=None):
        if scales is None:
            scales = [i / 10.0 for i in range(5, 2, 21)]

        all_matches = []
        max_score = -1
        best_scale = 1
        best_location = None
        for scale in scales:
            resized_template = cv2.resize(
                template, (int(template.shape[1] * scale), int(template.shape[0] * scale))
            )

            if (
                resized_template.shape[0] > image.shape[0]
                or resized_template.shape[1] > image.shape[1]
            ):
                continue

            result = cv2.matchTemplate(image, resized_template, cv2.TM_CCOEFF_NORMED)

            _, max_val, _, max_loc = cv2.minMaxLoc(result)

            if max_val > max_score:
                max_score = max_val
                best_scale = scale
                best_location = max_loc

        return best_scale

    @staticmethod
    def multi_scale_template_matching(
        image, template, threshold=0.9, scales=[i / 10.0 for i in range(5, 2, 21)]
    ):
        all_matches = []
        all_score = []
        all_scale = 1
        for scale in scales:
            # resized_template = cv2.resize(template, (int(template.shape[1] * scale), int(template.shape[0] * scale)))

            resized_template = cv2.resize(
                template,
                (
                    int(template.shape[1] * scale * all_scale),
                    int(template.shape[0] * scale * all_scale),
                ),
            )
            image = cv2.resize(
                image, (int(image.shape[1] * all_scale), int(image.shape[0] * all_scale))
            )

            if (
                resized_template.shape[0] > image.shape[0]
                or resized_template.shape[1] > image.shape[1]
            ):
                continue

            result = cv2.matchTemplate(image, resized_template, cv2.TM_CCOEFF_NORMED)

            # _, max_val, _, max_loc = cv2.minMaxLoc(result)

            locs = np.where(result >= threshold)
            for pt in zip(*locs[::-1]):  # Switch cols and rows
                all_matches.append((pt, scale))
                score_at_pt = result[pt[1], pt[0]]
                all_score.append(score_at_pt)

        return all_matches, all_score


    @staticmethod   
    def non_max_suppression(boxes, overlap_thresh, scores):
        boxes = np.array(boxes)

        if len(boxes) == 0:
            return [], []

        if boxes.dtype.kind == "i":
            boxes = boxes.astype("float")

        pick = []

        x1, y1, x2, y2 = boxes[:, 0], boxes[:, 1], boxes[:, 2], boxes[:, 3]

        area = (x2 - x1 + 1) * (y2 - y1 + 1)
        idxs = np.argsort(scores)

        while len(idxs) > 0:
            last = len(idxs) - 1
            i = idxs[last]
            pick.append(i)

            xx1 = np.maximum(x1[i], x1[idxs[:last]])
            yy1 = np.maximum(y1[i], y1[idxs[:last]])
            xx2 = np.minimum(x2[i], x2[idxs[:last]])
            yy2 = np.minimum(y2[i], y2[idxs[:last]])

            w = np.maximum(0, xx2 - xx1 + 1)
            h = np.maximum(0, yy2 - yy1 + 1)

            overlap = (w * h) / area[idxs[:last]]

            idxs = np.delete(
                idxs, np.concatenate(([last], np.where(overlap > overlap_thresh)[0]))
            )

        return boxes[pick].astype("int"), pick


    @staticmethod
    def filter_data(meta_data):
        return [entry for entry in meta_data if entry["properties"]["texts"][0]]

    @staticmethod
    def sort_data_by_y_coordinate(data):
        return sorted(data, key=lambda entry: entry["properties"]["rectangle"][1])

    @staticmethod
    def sort_row_by_x_coordinate(row):
        return sorted(row, key=lambda element: element["properties"]["rectangle"][0])

    def organize_elements(self, sorted_data):
        elements = []
        current_row = []
        prev_y = None

        for entry in sorted_data:
            y = entry["properties"]["rectangle"][1]

            if prev_y is not None and abs(y - prev_y) > 10:
                elements.append(self.sort_row_by_x_coordinate(current_row))
                current_row = []

            element_name = self.construct_element_name(entry)
            current_row.append(
                {"name": element_name, "rectangle": entry["properties"]["rectangle"]}
            )
            prev_y = y

        if current_row:
            elements.append(self.sort_row_by_x_coordinate(current_row))

        return elements

    @staticmethod
    def construct_element_name(entry):
        if entry["properties"]["friendly_class_name"] in ["MenuItem", "Button"]:
            return entry["properties"]["texts"][0]
        else:
            return f"{entry['properties']['texts'][0]}|{entry['properties']['friendly_class_name']}"

    def run_parallel_tasks(self, func, param_list):
        with ThreadPoolExecutor() as executor:
            futures = [executor.submit(func, *param) for param in param_list]
            return [future.result() for future in futures]

    @staticmethod
    def filter_data(meta_data):
        return [entry for entry in meta_data if entry["properties"]["texts"][0]]

    @staticmethod
    def sort_data_by_y_coordinate(data):
        return sorted(data, key=lambda entry: entry["properties"]["rectangle"][1])

    @staticmethod
    def sort_row_by_x_coordinate(row):
        return sorted(row, key=lambda element: element["properties"]["rectangle"][0])

    def organize_elements(self, sorted_data):
        elements = []
        current_row = []
        prev_y = None

        for entry in sorted_data:
            y = entry["properties"]["rectangle"][1]

            if prev_y is not None and abs(y - prev_y) > 10:
                elements.append(self.sort_row_by_x_coordinate(current_row))
                current_row = []

            element_name = self.construct_element_name(entry)
            current_row.append(
                {"name": element_name, "rectangle": entry["properties"]["rectangle"]}
            )
            prev_y = y

        if current_row:
            elements.append(self.sort_row_by_x_coordinate(current_row))

        return elements

    @staticmethod
    def construct_element_name(entry):
        if entry["properties"]["friendly_class_name"] in ["MenuItem", "Button"]:
            return entry["properties"]["texts"][0]
        else:
            return f"{entry['properties']['texts'][0]}|{entry['properties']['friendly_class_name']}"

    def run_parallel_tasks(self, func, param_list):
        with ThreadPoolExecutor() as executor:
            futures = [executor.submit(func, *param) for param in param_list]
            return [future.result() for future in futures]



import json
import asyncio
import platform
from collections.abc import Callable
from datetime import datetime
from enum import StrEnum
from typing import Any, cast, Dict, Callable

from anthropic import Anthropic, AnthropicBedrock, AnthropicVertex, APIResponse
from anthropic.types import TextBlock, ToolResultBlockParam
from anthropic.types.beta import BetaMessage, BetaTextBlock, BetaToolUseBlock, BetaMessageParam

from computer_use_ootb_internal.computer_use_demo.tools.screen_capture import get_screenshot
from computer_use_ootb_internal.computer_use_demo.gui_agent.llm_utils.oai import run_oai_interleaved, encode_image
from computer_use_ootb_internal.computer_use_demo.gui_agent.llm_utils.qwen import run_qwen
from computer_use_ootb_internal.computer_use_demo.gui_agent.llm_utils.llm_utils import extract_data
from computer_use_ootb_internal.computer_use_demo.tools.colorful_text import colorful_text_showui, colorful_text_vlm

import torch
from transformers import Qwen2VLForConditionalGeneration, AutoTokenizer, AutoProcessor
from qwen_vl_utils import process_vision_info

SYSTEM_PROMPT = f"""<SYSTEM_CAPABILITY>
* You are utilizing a Windows system with internet access.
* The current date is {datetime.today().strftime('%A, %B %d, %Y')}.
</SYSTEM_CAPABILITY>
"""

class LocalVLMPlanner:
    def __init__(
        self,
        model: str, 
        provider: str, 
        system_prompt_suffix: str, 
        output_callback: Callable, 
        api_response_callback: Callable,
        max_tokens: int = 4096,
        only_n_most_recent_images: int | None = None,
        selected_screen: int = 0,
        print_usage: bool = True,
        device: torch.device = torch.device("cpu"),
    ):
        self.device = device
        
        self.min_pixels = 256 * 28 * 28
        self.max_pixels = 1344 * 28 * 28
        
        if model == "qwen-vl-7b-instruct + ShowUI":  # local model
            self.model_name = "qwen-vl-7b-instruct"
            self.model = Qwen2VLForConditionalGeneration.from_pretrained(
                # "Qwen/Qwen2-VL-7B-Instruct",
                "./Qwen2-VL-7B-Instruct",
                torch_dtype=torch.bfloat16,
                device_map="cpu"
            ).to(self.device)
            self.processor = AutoProcessor.from_pretrained(
                "./Qwen2-VL-7B-Instruct",
                min_pixels=self.min_pixels,
                max_pixels=self.max_pixels
            )
        elif model == "qwen2-vl-2b-instruct":
            self.model_name = "qwen2-vl-2b-instruct"
            self.model = Qwen2VLForConditionalGeneration.from_pretrained(
                "./Qwen2-VL-2B-Instruct",
                torch_dtype=torch.bfloat16,
                device_map="cpu"
            ).to(self.device)
            self.processor = AutoProcessor.from_pretrained(
                "./Qwen2-VL-2B-Instruct",
                min_pixels=self.min_pixels,
                max_pixels=self.max_pixels
            )
        else:
            raise ValueError(f"Model {model} not supported")
        
        self.provider = provider
        self.system_prompt_suffix = system_prompt_suffix
        self.api_response_callback = api_response_callback
        self.max_tokens = max_tokens
        self.only_n_most_recent_images = only_n_most_recent_images
        self.selected_screen = selected_screen
        self.output_callback = output_callback
        self.system_prompt = self._get_system_prompt() + self.system_prompt_suffix

        self.print_usage = print_usage
        self.total_token_usage = 0
        self.total_cost = 0

           
    def __call__(self, messages: list):
        
        # drop looping actions msg, byte image etc
        planner_messages = _message_filter_callback(messages)  
        print(f"filtered_messages: {planner_messages}")

        # Take a screenshot
        screenshot, screenshot_path = get_screenshot(selected_screen=self.selected_screen)
        screenshot_path = str(screenshot_path)
        image_base64 = encode_image(screenshot_path)
        self.output_callback(f'Screenshot for {colorful_text_vlm}:\n<img src="data:image/png;base64,{image_base64}">',
                             sender="bot")
        
        if isinstance(planner_messages[-1], dict):
            if not isinstance(planner_messages[-1]["content"], list):
                planner_messages[-1]["content"] = [planner_messages[-1]["content"]]
            planner_messages[-1]["content"].append(screenshot_path)

        print(f"Sending messages to VLMPlanner: {planner_messages}")

        messages_for_processor = [
            {
                "role": "system",
                "content": [{"type": "text", "text": self.system_prompt}]
            },
            {
                "role": "user",
                "content": [
                {"type": "image", "image": screenshot_path, "min_pixels": self.min_pixels, "max_pixels": self.max_pixels},
                {"type": "text", "text": f"Task: {''.join(planner_messages)}"}
            ],
        }]
        
        text = self.processor.apply_chat_template(
            messages_for_processor, tokenize=False, add_generation_prompt=True
        )
        image_inputs, video_inputs = process_vision_info(messages_for_processor)
        
        inputs = self.processor(
            text=[text],
            images=image_inputs,
            videos=video_inputs,
            padding=True,
            return_tensors="pt",
        )
        inputs = inputs.to(self.device)

        generated_ids = self.model.generate(**inputs, max_new_tokens=128)
        generated_ids_trimmed = [
            out_ids[len(in_ids):] for in_ids, out_ids in zip(inputs.input_ids, generated_ids)
        ]
        vlm_response = self.processor.batch_decode(
            generated_ids_trimmed, skip_special_tokens=True, clean_up_tokenization_spaces=False
        )[0]
        
        print(f"VLMPlanner response: {vlm_response}")
        
        vlm_response_json = extract_data(vlm_response, "json")

        # vlm_plan_str = '\n'.join([f'{key}: {value}' for key, value in json.loads(response).items()])
        vlm_plan_str = ""
        for key, value in json.loads(vlm_response_json).items():
            if key == "Thinking":
                vlm_plan_str += f'{value}'
            else:
                vlm_plan_str += f'\n{key}: {value}'
        
        self.output_callback(f"{colorful_text_vlm}:\n{vlm_plan_str}", sender="bot")
        
        return vlm_response_json


    def _api_response_callback(self, response: APIResponse):
        self.api_response_callback(response)
        

    def reformat_messages(self, messages: list):
        pass

    def _get_system_prompt(self):
        os_name = platform.system()
        return f"""
You are using an {os_name} device.
You are able to use a mouse and keyboard to interact with the computer based on the given task and screenshot.
You can only interact with the desktop GUI (no terminal or application menu access).

You may be given some history plan and actions, this is the response from the previous loop.
You should carefully consider your plan base on the task, screenshot, and history actions.

Your available "Next Action" only include:
- ENTER: Press an enter key.
- ESCAPE: Press an ESCAPE key.
- INPUT: Input a string of text.
- CLICK: Describe the ui element to be clicked.
- HOVER: Describe the ui element to be hovered.
- SCROLL: Scroll the screen, you must specify up or down.
- PRESS: Describe the ui element to be pressed.


Output format:
```json
{{
    "Thinking": str, # describe your thoughts on how to achieve the task, choose one action from available actions at a time.
    "Next Action": "action_type, action description" | "None" # one action at a time, describe it in short and precisely. 
}}
```

One Example:
```json
{{  
    "Thinking": "I need to search and navigate to amazon.com.",
    "Next Action": "CLICK 'Search Google or type a URL'."
}}
```

IMPORTANT NOTES:
1. You should only give a single action at a time. for example, INPUT text, and ENTER can't be in one Next Action.
2. Attach the text to Next Action, if there is text or any description for the button. 
3. You should not include other actions, such as keyboard shortcuts.
4. When the task is completed, you should say "Next Action": "None" in the json field.
""" 

    

def _maybe_filter_to_n_most_recent_images(
    messages: list[BetaMessageParam],
    images_to_keep: int,
    min_removal_threshold: int = 10,
):
    """
    With the assumption that images are screenshots that are of diminishing value as
    the conversation progresses, remove all but the final `images_to_keep` tool_result
    images in place, with a chunk of min_removal_threshold to reduce the amount we
    break the implicit prompt cache.
    """
    if images_to_keep is None:
        return messages

    tool_result_blocks = cast(
        list[ToolResultBlockParam],
        [
            item
            for message in messages
            for item in (
                message["content"] if isinstance(message["content"], list) else []
            )
            if isinstance(item, dict) and item.get("type") == "tool_result"
        ],
    )

    total_images = sum(
        1
        for tool_result in tool_result_blocks
        for content in tool_result.get("content", [])
        if isinstance(content, dict) and content.get("type") == "image"
    )

    images_to_remove = total_images - images_to_keep
    # for better cache behavior, we want to remove in chunks
    images_to_remove -= images_to_remove % min_removal_threshold

    for tool_result in tool_result_blocks:
        if isinstance(tool_result.get("content"), list):
            new_content = []
            for content in tool_result.get("content", []):
                if isinstance(content, dict) and content.get("type") == "image":
                    if images_to_remove > 0:
                        images_to_remove -= 1
                        continue
                new_content.append(content)
            tool_result["content"] = new_content


def _message_filter_callback(messages):
    filtered_list = []
    try:
        for msg in messages:
            if msg.get('role') in ['user']:
                if not isinstance(msg["content"], list):
                    msg["content"] = [msg["content"]]
                if isinstance(msg["content"][0], TextBlock):
                    filtered_list.append(str(msg["content"][0].text))  # User message
                elif isinstance(msg["content"][0], str):
                    filtered_list.append(msg["content"][0])  # User message
                else:
                    print("[_message_filter_callback]: drop message", msg)
                    continue                

            # elif msg.get('role') in ['assistant']:
            #     if isinstance(msg["content"][0], TextBlock):
            #         msg["content"][0] = str(msg["content"][0].text)
            #     elif isinstance(msg["content"][0], BetaTextBlock):
            #         msg["content"][0] = str(msg["content"][0].text)
            #     elif isinstance(msg["content"][0], BetaToolUseBlock):
            #         msg["content"][0] = str(msg['content'][0].input)
            #     elif isinstance(msg["content"][0], Dict) and msg["content"][0]["content"][-1]["type"] == "image":
            #         msg["content"][0] = f'<img src="data:image/png;base64,{msg["content"][0]["content"][-1]["source"]["data"]}">'
            #     else:
            #         print("[_message_filter_callback]: drop message", msg)
            #         continue
            #     filtered_list.append(msg["content"][0])  # User message
                
            else:
                print("[_message_filter_callback]: drop message", msg)
                continue
            
    except Exception as e:
        print("[_message_filter_callback]: error", e)
                
    return filtered_list


import requests
import json

from computer_use_ootb_internal.computer_use_demo.gui_agent.planner.teach_mode_vlm_planner import TeachModeVLMPlanner, _message_filter_callback, _maybe_filter_to_n_most_recent_images
from computer_use_ootb_internal.computer_use_demo.gui_agent.llm_utils.oai import encode_image
from computer_use_ootb_internal.computer_use_demo.tools.screen_capture import get_screenshot
from computer_use_ootb_internal.computer_use_demo.gui_agent.llm_utils.llm_utils import extract_data


class RemoteTeachModeVLMPlanner(TeachModeVLMPlanner):
    _NAV_SYSTEM = """You are an assistant trained to navigate the desktop screen. 
    Given a task instruction, a screen observation, and an action history sequence, 
    output the next action and wait for the next observation. 
    Here is the action space:
    {_ACTION_SPACE}
    """
    
    _OOTB_Executor_Action_Space = """
        1. CLICK: Click on an element, value is not applicable and the position [x,y] is required. 
        2. INPUT: Type a string into an element, value is a string to type and the position [x,y] is required. 
        3. SELECT: Select a value for an element, value is not applicable and the position [x,y] is required. 
        4. HOVER: Hover on an element, value is not applicable and the position [x,y] is required.
        5. ANSWER: Answer the question, value is the answer and the position is not applicable.
        6. ENTER: Enter operation, value and position are not applicable.
        7. SCROLL: Scroll the screen, value is the direction to scroll and the position is not applicable.
        8. SELECT_TEXT: Select some text content, value is not applicable and position [[x1,y1], [x2,y2]] is the start and end position of the select operation.
        9. COPY: Copy the text, value is the text to copy and the position is not applicable.
        10. ESC: ESCAPE operation, value and position are not applicable.
        """
    
    
    
    def __init__(
        self,
        remote_api_url: str,  ##
        api_key: str,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.remote_api_url = remote_api_url
        self.api_key = api_key
        self.system_prompt = self._NAV_SYSTEM.format(
            _ACTION_SPACE=self._OOTB_Executor_Action_Space
        )

    def __call__(self, messages: list, screenshot_path: str = None):
        """
        Override the __call__ method to send data to the remote API.
        """
        in_context = self.history_manager.get_in_context_example(self.user_id, self.trace_id)

        # Filter and prepare messages
        planner_messages = _message_filter_callback(messages)

        if self.only_n_most_recent_images:
            _maybe_filter_to_n_most_recent_images(planner_messages, self.only_n_most_recent_images)

        # Take a screenshot if not provided
        if screenshot_path is None:
            _, screenshot_path = get_screenshot(selected_screen=self.selected_screen)

        # Encode the screenshot for sending
        image_base64 = encode_image(screenshot_path)

        # Add in-context example
        self.add_in_context_example(in_context)

        # Prepare the payload for the remote API
        payload = {
            "system_prompt": self.system_prompt,
            "messages": planner_messages,
            "screenshot": image_base64,
            "model": self.model,
            "max_tokens": self.max_tokens,
        }

        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }

        # Send the request to the remote API
        response = requests.post(self.remote_api_url, headers=headers, json=payload)

        if response.status_code != 200:
            print(f"Error: Received status code {response.status_code} from remote API.")
            print(f"Response: {response.text}")
            return None

        # Parse the API response
        response_data = response.json()

        # Extract the plan and token usage
        vlm_response_json = extract_data(response_data.get("vlm_response", ""), "json")
        token_usage = response_data.get("token_usage", 0)

        # Update total token usage and cost
        self.total_token_usage += token_usage
        self.total_cost += (token_usage * 0.15 / 1000000)

        print(f"{self.model} token usage: {token_usage}")
        print(f"VLMPlanner total token usage so far: {self.total_token_usage}. Total cost so far: $USD{self.total_cost:.5f}")

        # Format and display the plan
        vlm_plan_str = ""
        for key, value in json.loads(vlm_response_json).items():
            if key == "Thinking":
                vlm_plan_str += f'{value}'
            else:
                vlm_plan_str += f'\n{key}: {value}'

        return vlm_response_json

    def add_in_context_example(self, in_context_example: str):
        """
        Extend the add_in_context_example method to ensure it is sent to the remote API.
        """
        self.system_prompt += f"""NOTE: Reference the following action trajectory to do the task, when user ask you to do the similar task.    
IN-CONTEXT EXAMPLE:
{in_context_example}
"""

# Example usage
if __name__ == "__main__":
    remote_planner = RemoteTeachModeVLMPlanner(
        remote_api_url="https://example.com/api/vlm_planner",
        api_key="your_api_key_here",
        model="your_model_here",
        system_prompt_suffix="Custom prompt suffix here",
        output_callback=print,
        api_response_callback=print,
        user_id="user123",
        trace_id="trace123",
    )

    messages = [
        {"role": "user", "content": "Open the settings menu."}
    ]

    remote_planner(messages)



import cv2
import numpy as np
import datetime
import requests
import importlib
# import textdistance
import glob
from PIL import Image
import ctypes
from .gui_parser_adobe import AdobeParser

from .text_detection import text_detection
from .icon_detection.icon_detection import detect_icons, IconDetector
from .utils import *

import os

class GUIParser():
    name = "gui_parser"
    description = """
This tool can extract the information of screenshot.
Invoke command: gui_parser(query, visual[i])
:param query -> str, specific command. visual[i] -> image, the latest screenshot.
"""

    def __init__(self, cache_folder=".cache/", user_id=None, trace_id=None, scaleFactor="1.0x"):
        # judge if the cache folder exists
        os.makedirs(cache_folder, exist_ok=True)
        ocr_cache_folder = os.path.join(cache_folder, "ocr")
        os.makedirs(ocr_cache_folder, exist_ok=True)
        super(GUIParser, self).__init__()
        self.cache_folder = cache_folder
        print(f"Cache folder: {self.cache_folder}")
        
        self.task_id = get_current_time()
        self.parsers = {}
        self.temperature = 0
        self.gui_parser = None
        self.icon_template_threshold = 0.7
        try:
            self.ootb_database_path = os.path.join(os.getenv("OOTB_PATH"), "ootbdatabase")
        except Exception as e:
            print(f"Error: {e}. OOTB_PATH is not set correctly.")
        self.user_id = user_id
        self.trace_id = trace_id
        self.scaleFactor = scaleFactor
        self.width = None
        self.height = None

    def __call__(
        self,
        uia_data=None,
        screenshot_path=None,
        query=None,
        mode = "uia", # "uia", "ocr", or "teach"
        ocr_mode = "googleocr"
    ):
        '''
        query: list, represent one step of the task. [text_query, icon_path, text_query, icon_path, ...]
        '''
        image = Image.open(screenshot_path)
        self.width, self.height = image.size
    
        if mode == "teach":
            # self.parsed_gui = {"Screen": None}
            ocr = text_detection(screenshot_path, ocr_mode)
            icon_items = []
            
            print("query: ", query)
            # 处理图标检测
            if query and isinstance(query, list) and len(query) > 1:
                specific_icon_names = [item for item in query if self.is_image_path(item)]
                if specific_icon_names:
                    icon_template_folder = os.path.join(self.ootb_database_path, self.user_id, self.trace_id, "icons")

                    print(f"icon_template_folder: {icon_template_folder}")

                    icon_detector = IconDetector(icon_template_folder)

                    detected_items = icon_detector(
                        screenshot_path=screenshot_path,
                        mode="teach",
                        threshold=self.icon_template_threshold, # TODO: setting on init
                        scale_factor=self.scaleFactor,
                        specific_icon_names=specific_icon_names,
                    )
                    
                    if detected_items:
                        icon_items = detected_items
            print("detected_icons:", icon_items)
            # 获取OCR文本结果
            ocr_elements = []
            if ocr and "texts" in ocr:
                for text_item in ocr["texts"]:
                    ocr_elements.append({
                        "name": text_item["content"],
                        "rectangle": text_item["bbox"],
                        "is_text": True
                    })
            
            # 合并所有元素
            all_elements = []
            all_elements.extend(icon_items)  # 添加图标元素
            all_elements.extend(ocr_elements)  # 添加文本元素
            
            # 创建主面板
            ocr_icon_parsed_gui = {
                "Screen": [{
                    "name": "Main Content",
                    "rectangle": [0, 0, self.width, self.height],
                    "elements": self.merge_elements({"combined_elements": all_elements})
                }]
            }
            
            # get uia parsed gui
            uia_parsed_gui = self.get_panel_uia(uia_data)
            self.postprocess_uia(uia_parsed_gui)
            
            # merge the code
            self.parsed_gui = self.merge_uia_and_teach_results(uia_parsed_gui, ocr_icon_parsed_gui)
            return self.parsed_gui
        
        # only use uia mode
        elif mode == "uia":
            # Parse the UI elements of the current window
            self.parsed_gui = {}
            self.exclude_class_name_list = []
            # if self.if_browser(window_name):
            #     self.exclude_class_name_list = [
            #         "Custom",
            #         "Menu",
            #         "Pane",
            #         "TabControl",
            #         "DataItem",
            #     ]
            # elif self.if_office(window_name):
            #     self.exclude_class_name_list = [
            #         "Custom",
            #         "Menu",
            #         "Pane",
            #         "Toolbar",
            #         "TabControl",
            #         "TreeItem",
            #         "DataItem",
            #         "Hyperlink",
            #     ]
            # else:
            #     self.exclude_class_name_list = []

            self.parsed_gui = self.get_panel_uia(uia_data)
            self.postprocess_uia(self.parsed_gui)

            return self.parsed_gui
        
        # OCR + full icon detection mode
        elif mode == "icon_ocr": 
            ocr = text_detection(screenshot_path)
            icon_template_folder = os.path.join(self.ootb_database_path, "ootbdatabase/teach_mode")
            icon_detector = IconDetector(icon_template_folder)
            icon_items = [
                icon_detector(
                    screenshot_path=screenshot_path,  mode="full_detect", threshold=0.6, scale_factor=self.scaleFactor
                )
            ]
            self.parsed_gui = self.postprocess_icon(
                "screen", icon_items, screenshot_path
            )
            for panel_item in self.parsed_gui["screen"]:
                temp = {}
                temp["editing_control"] = self.get_text(
                    panel_item, ocr, screenshot_path
                )

                panel_item["elements"] += self.merge_elements(temp)
                
        # OCR only mode
        else: 
            ocr = text_detection(screenshot_path)
            self.parsed_gui = self.postprocess_icon(
                "screen", [], screenshot_path
            )
            for panel_item in self.parsed_gui["screen"]:
                temp = {}
                temp["editing_control"] = self.get_text(
                    panel_item, ocr, screenshot_path
                )

                panel_item["elements"] += self.merge_elements(temp)
        
            
            return self.parsed_gui



    def if_browser(self, window_name):
        # check if ["Edge", "Chrome", "Firefox", "Safari", "Opera"] in window_name
        if any(browser in window_name for browser in ["Edge", "Chrome", "Firefox", "Safari", "Opera"]):
            return True
        return False
    
    def if_adobe(self, window_name):
        # check if ["Adobe Premiere Pro", "Adobe After Effect", "Adobe Acrobat"] in window_name
        if any(adobe in window_name for adobe in ["Adobe Premiere Pro", "Adobe After Effect", "Adobe Acrobat"]):
            return True
        return False

    def if_office(self, window_name):
        # check if ["Excel", "Word", "PowerPoint", "Outlook"] in window_name
        if any(office in window_name for office in ["Excel", "Word", "PowerPoint", "Outlook"]):
            return True
        return False
    
    @staticmethod
    def is_image_path(text):
        # Checking if the input text ends with typical image file extensions
        image_extensions = (".jpg", ".jpeg", ".png", ".gif", ".bmp", ".tiff", ".tif")
        if text.endswith(image_extensions):
            return True
        else:
            return False

    @staticmethod
    def get_text(panel_item, ocr, screenshot_path):
        # Step1: Find all the texts in the panel
        panel_rect = panel_item["rectangle"]
        # 过滤出在panel区域内的文本
        panel_texts = []
        for item in ocr["texts"]:
            bbox = item["bbox"]
            # 检查文本是否在panel区域内
            if (bbox[0] >= panel_rect[0] and 
                bbox[1] >= panel_rect[1] and 
                bbox[2] <= panel_rect[2] and 
                bbox[3] <= panel_rect[3]):
                panel_texts.append(item)

        if not panel_texts:
            return []
            
        # obtain information about panel texts
        sorted_panel_texts = sorted(
            panel_texts, key=lambda x: (x["bbox"][1], x["bbox"][0])
        )
        
        # Step 2: Identify rows by grouping elements that are in approximately the same vertical position.
        editing_controls = []
        current_row = []
        if sorted_panel_texts:
            previous_y = sorted_panel_texts[0]["bbox"][1]

            for item in sorted_panel_texts:
                y1 = item["bbox"][1]
                # If the vertical position of the current item is significantly different from the previous one, start a new row.
                if abs(y1 - previous_y) > 15:
                    if current_row:  # 只有当current_row非空时才添加
                        editing_controls.append(current_row)
                    current_row = []

                current_row.append(
                    {
                        "name": item["content"],
                        "rectangle": item["bbox"],
                    }
                )
                previous_y = y1
                
            # Add the last row if not empty.
            if current_row:
                editing_controls.append(current_row)
                
        # Step 3: Sort elements within each row based on their horizontal position.
        for i, row in enumerate(editing_controls):
            editing_controls[i] = sorted(row, key=lambda x: x["rectangle"][0])
            
        return editing_controls
        
    def postprocess_icon(self, window_name, icon_items, screenshot_path):
        image = Image.open(screenshot_path)
        width, height = image.size
        result = {
            window_name: [
                {
                    "name": "Main Content",
                    "rectangle": [0, 0, width, height],
                    "elements": None
                }
            ]
        }
        return result

    @staticmethod
    def merge_elements(panel_item):
        # 检查是否为空
        if not panel_item or not any(panel_item.values()):
            return []
        
        # 收集所有元素
        all_elements = []
        for key, value in panel_item.items():
            if isinstance(value, list):
                all_elements.extend(value)
        
        if not all_elements:
            return []
        
        # 按Y轴分组,同一行的阈值为15像素
        y_sorted = sorted(all_elements, key=lambda x: (x["rectangle"][1] + x["rectangle"][3]) / 2)
        merged_control = []
        current_row = []
        previous_y = None
        
        for element in y_sorted:
            y_center = (element["rectangle"][1] + element["rectangle"][3]) / 2
            
            if previous_y is not None and abs(y_center - previous_y) > 15:
                if current_row:
                    # 对当前行按X轴排序
                    current_row.sort(key=lambda x: x["rectangle"][0])
                    merged_control.append(current_row)
                current_row = []
                
            current_row.append(element)
            previous_y = y_center
        
        # 添加最后一行
        if current_row:
            current_row.sort(key=lambda x: x["rectangle"][0])
            merged_control.append(current_row)
            
        # 对所有行按Y轴中心点排序
        merged_control.sort(
            key=lambda row: (row[0]["rectangle"][1] + row[0]["rectangle"][3]) / 2
        )
        
        return merged_control

    def postprocess_uia(self, metadata):
        # iterate each window
        for window_data in metadata.values():

            for element_data in window_data:
                # get all elements of current level
                elements = element_data.get("elements", [])

                # if the class name is TitleBar, regain the rectangle of the whole panel
                if element_data["class_name"] == "TitleBar" and len(elements) > 0:
                    left = min(e.get("rectangle", [0, 0, 0, 0])[0] for e in elements)
                    top = min(e.get("rectangle", [0, 0, 0, 0])[1] for e in elements)
                    right = max(e.get("rectangle", [0, 0, 0, 0])[2] for e in elements)
                    bottom = max(e.get("rectangle", [0, 0, 0, 0])[3] for e in elements)
                    element_data["rectangle"] = [left, top, right, bottom]
                # sort elements by x and y
                element_data["elements"] = sort_elements_by_xy(elements)

        return metadata
    
    def merge_uia_and_teach_results(self,uia_result, teach_result):
        def calculate_overlap(rect1, rect2):
            # 计算两个矩形的重叠面积
            x1 = max(rect1[0], rect2[0])
            y1 = max(rect1[1], rect2[1])
            x2 = min(rect1[2], rect2[2])
            y2 = min(rect1[3], rect2[3])
            
            if x1 >= x2 or y1 >= y2:
                return 0
                
            overlap_width = x2 - x1
            overlap_height = y2 - y1
            
            return overlap_width * overlap_height

        def is_significant_overlap(ocr_element, uia_element):
            ocr_rect = ocr_element["rectangle"]
            uia_rect = uia_element["rectangle"]
            
            overlap_area = calculate_overlap(ocr_rect, uia_rect)
            ocr_area = (ocr_rect[2] - ocr_rect[0]) * (ocr_rect[3] - ocr_rect[1])
            
            return overlap_area > 0.5 * ocr_area if ocr_area > 0 else False

        merged_result = uia_result.copy()

        all_ocr_elements = []
        # 遍历teach模式下的OCR文本元素
        for screen in teach_result.get("Screen", []):
            for elements_row in screen.get("elements", []):
                for element in elements_row:
                    if element["is_text"]:
                        should_add = True
                        # 检查是否与任何UIA元素有显著重叠
                        for window_name, panels in merged_result.items():
                            for panel in panels:
                                for elements_list in panel.get("elements", []):
                                    for uia_element in elements_list:
                                        if is_significant_overlap(element, uia_element):
                                            should_add = False
                                            break
                                    if not should_add:
                                        break
                                if not should_add:
                                    break
                            if not should_add:
                                break
                                
                        # 如果没有显著重叠，将OCR元素添加到最合适的面板中
                        if should_add:
                            # 找到最合适的面板（这里简单地添加到第一个面板）
                            all_ocr_elements.append(element)
                    else:
                        all_ocr_elements.append(element)
                        
        merged_result["OCR"] = [{
                    "name": "Main Content",
                    "rectangle": [0, 0, self.width, self.height],
                    "elements": [all_ocr_elements]
                }]
        return merged_result
    def get_panel_uia(self, control_info_list):
        # 定义文本到面板名称的映射
        text2panel_name = {
            "新建会话属性": "New session properties",
            "会话管理器": "Session manager",
            "任务栏": "Taskbar",
        }

        # 递归处理控件信息
        def recurse_controls(control_info, dialog_components, depth, software_name):
            children = control_info.get("children", [])
            if not children:
                return
            
            for child_control in children:
                child_properties = child_control.get("properties", {})
                child_friendly_class_name = child_properties.get("friendly_class_name", "")
                
                # 如果控件类型在排除列表中，跳过处理
                if child_friendly_class_name in self.exclude_class_name_list:
                    continue
                    
                # 获取控件名称
                child_texts = child_properties.get("texts", [])
                if not child_texts:
                    child_properties_name = ""
                elif isinstance(child_texts[0], list):
                    result = []
                    for item in child_texts:
                        if isinstance(item, list) and item and isinstance(item[0], str):
                            result.append("".join(item))
                        elif isinstance(item, str):
                            result.append(item)
                    child_properties_name = "".join(result)
                else:
                    child_properties_name = child_texts[0]
                
                # 特殊处理编辑框和组合框
                if child_friendly_class_name in ["Edit", "ComboBox"] and not child_properties_name:
                    child_properties_name = "Search Bar"
                
                # 判断是否需要处理该控件
                rectangle = child_properties.get("rectangle", [0, 0, 0, 0])
                automation_id = child_properties.get("automation_id", "")
                
                # 放宽控件过滤条件
                browser_list = ["Edge", "Chrome", "Firefox", "Opera", "Safari"]
                is_browser = False
                for browser in browser_list:
                    if browser in software_name:
                        is_browser = True
                        break
                should_process = (
                    # 任务栏特殊处理
                    (software_name == "Taskbar" and child_friendly_class_name in ["Button", "Static", "Image"]) or
                    # 浏览器特殊处理
                    (is_browser and child_friendly_class_name in ["Pane", "Button", "Edit", "Tab", "TabItem", "Document", "Group"]) or
                    # 常规控件处理
                    (not child_control.get("children", []) or  # 没有子控件的元素保留
                     automation_id or  # 有automation_id的控件保留
                     child_friendly_class_name not in self.exclude_class_name_list)  # 不在排除列表中的控件保留
                ) and (
                    # 基本条件检查
                    len(rectangle) == 4 and
                    not all(element == 0 for element in rectangle) and
                    (child_properties_name not in ["", '"'] or automation_id)  # 允许有automation_id的空名称控件
                )
                
                if should_process:
                    # 获取控件的矩形边界
                    left, top, right, bottom = rectangle
                    dialog_rect = dialog_components.get("rectangle", [0, 0, 0, 0])
                    if len(dialog_rect) != 4:
                        dialog_rect = [0, 0, 0, 0]
                    left_bound, top_bound, right_bound, bottom_bound = dialog_rect

                    # 根据软件名称进行特殊逻辑处理
                    if software_name not in ["Taskbar"] and not is_browser:  # 任务栏和浏览器不裁剪坐标
                        child_properties["rectangle"] = [
                            max(left, left_bound),
                            max(top, top_bound),
                            min(right, right_bound),
                            min(bottom, bottom_bound)
                        ]
                    else:
                        child_properties["rectangle"] = [left, top, right, bottom]
                    
                    # 检查矩形有效性
                    if (child_properties["rectangle"][0] < child_properties["rectangle"][2] 
                        and child_properties["rectangle"][1] < child_properties["rectangle"][3]):
                        
                        # 添加控件信息到对话框组件
                        element_info = {
                            "name": child_properties_name.replace("\u200b", ""),
                            "rectangle": child_properties["rectangle"],
                            "class_name": child_friendly_class_name,
                            "type": ["Click", "rightClick"],
                            "depth": f"{depth}-{dialog_components.get('count', 1)}"
                        }
                        
                        # 添加automation_id（如果存在）
                        if automation_id:
                            element_info["automation_id"] = automation_id
                            
                        dialog_components.setdefault("elements", []).append(element_info)
                        dialog_components["count"] = dialog_components.get("count", 1) + 1

                # 递归处理子控件
                recurse_controls(child_control, dialog_components, depth, software_name)

        dialog_components = {}

        # 遍历每个软件的控件信息
        for software_name, controls in control_info_list.items():
            if not controls:  # 跳过空的控件列表
                continue
                
            dialog_components[software_name] = []
            
            for control_info in controls:
                # 判断是否为需要处理的对话框类型
                control_properties = control_info.get("properties", {})
                friendly_class_name = control_properties.get("friendly_class_name", "")
                
                # 放宽面板类型的限制
                panel_types = ["Dialog", "Pane", "GroupBox", "TitleBar", "Menu", "Document", "ListBox", "AppBar", "Tab", "Group"]
                if (friendly_class_name in panel_types and control_info.get("children")):
                    
                    control_texts = control_properties.get("texts", [""])
                    control_name = control_texts[0] if control_texts else ""
                    automation_id = control_properties.get("automation_id", "")
                    
                    # 设置控件名称
                    if not control_name and automation_id:
                        control_name = automation_id  # 使用automation_id作为备选名称
                    elif not control_name:
                        if friendly_class_name == "TitleBar":
                            control_name = "Title Bar"
                        elif friendly_class_name == "Document":
                            control_name = "Main Content"
                        elif friendly_class_name == "Pane":
                            if software_name == "Taskbar":
                                control_name = "Taskbar"
                            elif "Edge" in software_name or "Chrome" in software_name or "Firefox" in software_name:
                                control_name = "Browser Content"
                            else:
                                control_name = "Navigation Bar" if software_name in ["web", "web video"] else "Main Content"
                    elif friendly_class_name == "Document" and software_name in ["web", "web video"]:
                        control_name = "Main Content"
                    
                    # 转换文本名称
                    control_name = text2panel_name.get(control_name, control_name)

                    # 初始化对话框组件并递归处理控件
                    panel_info = {
                        "name": control_name,
                        "rectangle": control_properties.get("rectangle", [0, 0, 0, 0]),
                        "class_name": friendly_class_name,
                        "depth": "1",
                        "elements": [],
                        "count": 1
                    }
                    
                    # 添加automation_id（如果存在）
                    if automation_id:
                        panel_info["automation_id"] = automation_id
                        
                    dialog_components[software_name].append(panel_info)
                    recurse_controls(control_info, dialog_components[software_name][-1], "1", software_name)
                    
        return dialog_components
    
import matplotlib.pyplot as plt
import cv2
import matplotlib.font_manager as fm

def show_img_with_box_points(img, box, ax, color="red"):
    x0, y0 = box[0], box[1]
    w, h = box[2] - box[0], box[3] - box[1]
    ax.add_patch(plt.Rectangle((x0, y0), w, h, edgecolor=color, facecolor=(0, 0, 0, 0), lw=0.4))

def extract_second_underscore_content(filename):
    parts = filename.split('_')
    if len(parts) > 1 and "@" in filename:
        return parts[1]
    return filename

def show_guiParserd(gui, img_path):
    img = cv2.imread(img_path)
    img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    plt.figure(dpi=500)
    plt.axis('off')
    plt.imshow(img)

    # 设置微软雅黑字体
    # font_path = "C:/Windows/Fonts/msyh.ttc"  # Windows系统上的微软雅黑字体路径
    # prop = fm.FontProperties(fname=font_path)
    item_cache = []
    for windows_name, panel_list in gui.items():
        print("windows_name: ", windows_name)
        print("panel_list: ", panel_list)
        for sequence in panel_list:
            # print("sequence: ", sequence)
            show_img_with_box_points(img, sequence["rectangle"], plt.gca(), color="green")
            if len(sequence["name"]) < 100:
                plt.text(sequence["rectangle"][0], sequence["rectangle"][1], sequence["name"], 
                        fontsize=4, color='green')
            for _, items in enumerate(sequence['elements']):
                item_cache.extend(items)
    for item in item_cache:
        try:
            if "rectangle" in item:
                if len(item["name"]) > 200:
                    continue
                box = item["rectangle"]
                if "current_step_icon" in item.keys() and item["current_step_icon"]:
                    show_img_with_box_points(img, box, plt.gca(),color="green")
                    plt.text(box[0], box[3]+25, "Current Step Icon", fontsize=3, color='green')
                else:
                    show_img_with_box_points(img, box, plt.gca(),color="blue")
                plt.text(box[0], box[1], extract_second_underscore_content(item["name"]), 
                        fontsize=2, color='red')
        except Exception as e:
            print(e)
    # plt.savefig('gui_parser5.png', bbox_inches='tight', pad_inches=0)
    plt.show()



def get_screen_info():
    try:
        scaleFactor = ctypes.windll.shcore.GetScaleFactorForDevice(0) / 100
        scaleFactor = str(scaleFactor) + "x"
    except Exception as e:
        print(e)
        scaleFactor = "1.0x"
    return scaleFactor
        
def parse_gui(user_id, trace_id, screenshot_path, user_scaleFactor="auto", uia_data=None, query=None, mode="uia", ocr_mode="googleocr"):
    # TODO: check mult-screens
    if user_scaleFactor == "auto":
        user_scaleFactor = get_screen_info()
    print("user_scaleFactor: ", user_scaleFactor)
    gui_parser = GUIParser(user_id=user_id, trace_id=trace_id, scaleFactor=user_scaleFactor)
    return gui_parser(uia_data=uia_data, screenshot_path=screenshot_path, query=query, mode=mode, ocr_mode=ocr_mode)

import json
import sys
if __name__ == "__main__":

    screenshot_path = r'D:\develop\computer_use_ootb_internal-main\.cache\20241222_041910\screenshot-0.png'
    meta_data = json.load(open("test_uia_data.json", "r"))
    gui_parser = GUIParser(user_id="test", trace_id="test", scaleFactor="1.0x")
    uia_gui = gui_parser.get_panel_uia(meta_data)
    uia_gui = gui_parser.postprocess_uia(uia_gui)
    with open("test_uia_gui.json", "w") as f:
        json.dump(uia_gui, f, indent=4)






import os
import logging
import base64
import requests

# from computer_use_ootb_internal.computer_use_demo.gui_agent.llm_utils import is_image_path, encode_image

def is_image_path(text):
    # Checking if the input text ends with typical image file extensions
    image_extensions = (".jpg", ".jpeg", ".png", ".gif", ".bmp", ".tiff", ".tif")
    if text.endswith(image_extensions):
        return True
    else:
        return False

def encode_image(image_path):
    """Encode image file to base64."""
    with open(image_path, "rb") as image_file:
        return base64.b64encode(image_file.read()).decode("utf-8")
    


# from openai import OpenAI
# client = OpenAI(
#     api_key=os.environ.get("OPENAI_API_KEY")
#     )


def run_oai_interleaved(messages: list, system: str, llm: str, api_key: str=None, max_tokens=256, temperature=0):
    """
    messages: example: [{"role": "user", "content": [{"type": "text", "text": "What is in the screenshot?"}, {"type": "image_url", "image_url": {"url": "data:image/jpeg;base64,..."}}]}
    
    """


    api_key = api_key or os.environ.get("OPENAI_API_KEY")
    if not api_key:
        raise ValueError("OPENAI_API_KEY is not set")
    
    headers = {"Content-Type": "application/json",
               "Authorization": f"Bearer {api_key}"}

    final_messages = [{"role": "system", "content": system}]

    # image_url = "https://upload.wikimedia.org/wikipedia/commons/thumb/d/dd/Gfp-wisconsin-madison-the-nature-boardwalk.jpg/2560px-Gfp-wisconsin-madison-the-nature-boardwalk.jpg"
    if type(messages) == list:
        for item in messages:
            contents = []
            if isinstance(item, dict):
                for cnt in item["content"]:
                    if isinstance(cnt, str):
                        print(f"cnt: {cnt}")
                        if is_image_path(cnt):
                            print(f"is_image_path: {cnt}")
                            base64_image = encode_image(cnt)
                            content = {"type": "image_url", "image_url": {"url": f"data:image/jpeg;base64,{base64_image}"}}
                        # content = {"type": "image_url", "image_url": {"url": image_url}}
                        else:
                            print(f"is_text: {cnt}")
                            content = {"type": "text", "text": cnt}
                    contents.append(content)
                    
                message = {"role": item["role"], "content": contents}
            else:  # str
                contents.append({"type": "text", "text": item})
                message = {"role": "user", "content": contents}
            
            final_messages.append(message)

    
    elif isinstance(messages, str):
        final_messages = [{"role": "user", "content": messages}]

    print("[oai] sending messages:", final_messages)

    payload = {
        "model": llm,
        "messages": final_messages,
        "max_tokens": max_tokens,
        "temperature": temperature,
        # "stop": stop,
    }

    # from IPython.core.debugger import Pdb; Pdb().set_trace()

    response = requests.post(
        "https://api.openai.com/v1/chat/completions", headers=headers, json=payload
    )

    try:
        text = response.json()['choices'][0]['message']['content']
        token_usage = int(response.json()['usage']['total_tokens'])
        return text, token_usage
        
    # return error message if the response is not successful
    except Exception as e:
        print(f"Error in interleaved openAI: {e}. This may due to your invalid OPENAI_API_KEY. Please check the response: {response.json()} ")
        return response.json()


if __name__ == "__main__":
    
    api_key = os.environ.get("OPENAI_API_KEY")
    if not api_key:
        raise ValueError("OPENAI_API_KEY is not set")
    
    text, token_usage = run_oai_interleaved(
        messages= [{"content": [
                        "What is in the screenshot?",   
                        "./tmp/outputs/screenshot_0b04acbb783d4706bc93873d17ba8c05.png"],
                    "role": "user"
                    }],
        llm="gpt-4o-mini",
        system="You are a helpful assistant",
        api_key=api_key,
        max_tokens=256,
        temperature=0)
    
    print(text, token_usage)
    # There is an introduction describing the Calyx... 36986



import os
import re
import json
from typing import List, Dict, Optional

class HistoryManager:
    def __init__(self, base_path: str = r"./computer_use_demo/ootbdatabase"):
        """
        初始化历史记录管理器
        
        Args:
            base_path (str): 基础路径，默认为'data'
        """
        self.base_path = base_path
        
    def load_trace(self, user_id: str, trace_id: str) -> Optional[Dict]:
        """
        加载指定用户ID和轨迹ID的轨迹信息
        
        Args:
            user_id (str): 用户ID
            trace_id (str): 轨迹ID
            
        Returns:
            Dict: 轨迹信息字典，如果文件不存在则返回None
        """
        file_path = os.path.join(self.base_path, user_id, trace_id, "trace_information.json")
        
        try:
            with open(file_path, 'r', encoding='utf-8') as f:
                trace_data = json.load(f)
            return trace_data
        except FileNotFoundError:
            print(f"找不到轨迹文件: {file_path}")
            return None
        except json.JSONDecodeError:
            print(f"JSON解析错误: {file_path}")
            return None
            
    def get_user_traces(self, user_id: str) -> List[str]:
        """
        获取指定用户的所有轨迹ID
        
        Args:
            user_id (str): 用户ID
            
        Returns:
            List[str]: 轨迹ID列表
        """
        user_path = os.path.join(self.base_path, user_id)
        if not os.path.exists(user_path):
            return []
            
        return [d for d in os.listdir(user_path) 
                if os.path.isdir(os.path.join(user_path, d))]
                
    def get_all_users(self) -> List[str]:
        """
        获取所有用户ID
        
        Returns:
            List[str]: 用户ID列表
        """
        if not os.path.exists(self.base_path):
            return []
            
        return [d for d in os.listdir(self.base_path) 
                if os.path.isdir(os.path.join(self.base_path, d))]
                
    def get_trace_summary(self, user_id: str, trace_id: str) -> Optional[Dict]:
        """
        获取轨迹的摘要信息
        
        Args:
            user_id (str): 用户ID
            trace_id (str): 轨迹ID
            
        Returns:
            Dict: 包含任务名称、描述和ID的字典
        """
        trace_data = self.load_trace(user_id, trace_id)
        if not trace_data:
            return None
            
        return {
            "taskName": trace_data.get("taskName"),
            "taskDescription": trace_data.get("taskDescription"),
            "taskId": trace_data.get("taskId")
        }
        
    def get_trajectory_actions(self, user_id: str, trace_id: str) -> Optional[List[Dict]]:
        """
        获取轨迹中的所有动作
        
        Args:
            user_id (str): 用户ID
            trace_id (str): 轨迹ID
            
        Returns:
            List[Dict]: 动作列表
        """
        trace_data = self.load_trace(user_id, trace_id)
        if not trace_data:
            return None
            
        return trace_data.get("trajectory", [])
    
    def get_in_context_example(self, user_id: str, trace_id: str) -> str:
        """
        Get the in-context example for the given user and trace
        """
        trace_data = self.load_trace(user_id, trace_id)
        if not trace_data:
            return None
        
        # only keep the action description
        steps = trace_data.get("trajectory", [])
        output = []

        for action in steps:
            try:
                # print(action["action_discription"])
                output.append([re.split(r'[\\/]', text)[-1] for text in action["action_discription"]])

            except Exception as e:
                from IPython.core.debugger import set_trace
                set_trace()

        return self.format_json(output) 
    
    def format_json(self, data):
        """
        格式化 JSON 数据，去掉 [ 和 ]，并将每组元素显示在一行。
        """
        result = []
        for ix, group in enumerate(data):
            line = f"{ix+1}: {'| '.join(map(str, group))}"  # 将每组元素转为字符串后用逗号分隔
            result.append(line)
        return "\n".join(result)  # 每组元素作为一行




        
        



import ast
import asyncio
from typing import Any, Dict, cast, List, Union
from collections.abc import Callable
import uuid
from anthropic.types.beta import (
    BetaContentBlock,
    BetaContentBlockParam,
    BetaImageBlockParam,
    BetaMessage,
    BetaMessageParam,
    BetaTextBlockParam,
    BetaToolResultBlockParam,
)
from anthropic.types import TextBlock
from anthropic.types.beta import BetaMessage, BetaTextBlock, BetaToolUseBlock
from computer_use_ootb_internal.computer_use_demo.tools import BashTool, ComputerTool, EditTool, ToolCollection, ToolResult
from computer_use_ootb_internal.computer_use_demo.tools.colorful_text import colorful_text_showui, colorful_text_vlm


class ShowUIExecutor:
    def __init__(
        self, 
        output_callback: Callable[[BetaContentBlockParam], None], 
        tool_output_callback: Callable[[Any, str], None],
        selected_screen: int = 0
    ):
        self.output_callback = output_callback
        self.tool_output_callback = tool_output_callback
        self.selected_screen = selected_screen
        self.screen_bbox = self._get_screen_resolution()
        print("Screen BBox:", self.screen_bbox)
        
        self.tool_collection = ToolCollection(
            ComputerTool(selected_screen=selected_screen, is_scaling=False),
            BashTool(),
            EditTool(),
        )
        
        self.supported_action_type={
            # "showui_action": "anthropic_tool_action"
            "CLICK": 'key',  # TBD
            "INPUT": "key",
            "ENTER": "key",  # TBD
            "ESC": "key",
            "ESCAPE": "key",
            "PRESS":  "key",
        }

    def __call__(self, response: str, messages: list[BetaMessageParam]):
        # response is expected to be :
        # {'content': "{'action': 'CLICK', 'value': None, 'position': [0.83, 0.15]}, ...", 'role': 'assistant'}, 
        
        action_dict = self._format_actor_output(response)  # str -> dict
        
        actions = action_dict["content"]
        role = action_dict["role"]
        
        # Parse the actions from showui
        action_list = self._parse_showui_output(actions)
        print("Parsed Action List:", action_list)
        
        tool_result_content = None
        
        if action_list is not None and len(action_list) > 0:
                    
            for action in action_list:  # Execute the tool (adapting the code from anthropic_executor.py)
            
                tool_result_content: list[BetaToolResultBlockParam] = []
                
                self.output_callback(f"{colorful_text_showui}:\n{action}", sender="bot")
                print("Converted Action:", action)
                
                sim_content_block = BetaToolUseBlock(id=f'toolu_{uuid.uuid4()}',
                                        input={'action': action["action"], 'text': action["text"], 'coordinate': action["coordinate"]},
                                        name='computer', type='tool_use')
                
                # update messages
                new_message = {
                    "role": "assistant",
                    "content": cast(list[BetaContentBlockParam], [sim_content_block]),
                }
                if new_message not in messages:
                    messages.append(new_message)

                # Run the asynchronous tool execution in a synchronous context
                result = asyncio.run(self.tool_collection.run(
                    name=sim_content_block.name,
                    tool_input=cast(dict[str, Any], sim_content_block.input),
                ))
                
                tool_result_content.append(
                    _make_api_tool_result(result, sim_content_block.id)
                )
                # print(f"executor: tool_result_content: {tool_result_content}")
                self.tool_output_callback(result, sim_content_block.id)

                # Craft messages based on the content_block
                # Note: to display the messages in the gradio, you should organize the messages in the following way (user message, bot message)
                display_messages = _message_display_callback(messages)
                # Send the messages to the gradio
                for user_msg, bot_msg in display_messages:
                    yield [user_msg, bot_msg], tool_result_content

        # if not tool_result_content:
        #     print(f"executor: No tool result content. Return {messages}.")
        #     return messages
        
        return tool_result_content
    
    
    def _format_actor_output(self, action_output: str|dict) -> Dict[str, Any]:
        if type(action_output) == dict:
            return action_output
        else:
            try:
                action_output.replace("'", "\"")
                action_dict = ast.literal_eval(action_output)
                return action_dict
            except Exception as e:
                print(f"Error parsing action output: {e}")
                return None
    

    def _parse_showui_output(self, output_text: str) -> Union[List[Dict[str, Any]], None]:
        try:
            output_text = output_text.strip()
            
            # process single dictionary
            if output_text.startswith("{") and output_text.endswith("}"):
                output_text = f"[{output_text}]"

            # Validate if the output resembles a list of dictionaries
            if not (output_text.startswith("[") and output_text.endswith("]")):
                raise ValueError("Output does not look like a valid list or dictionary.")

            print("Output Text:", output_text)

            parsed_output = ast.literal_eval(output_text)

            print("Parsed Output:", parsed_output)

            if isinstance(parsed_output, dict):
                parsed_output = [parsed_output]
            elif not isinstance(parsed_output, list):
                raise ValueError("Parsed output is neither a dictionary nor a list.")

            if not all(isinstance(item, dict) for item in parsed_output):
                raise ValueError("Not all items in the parsed output are dictionaries.")

            # refine key: value pairs, mapping to the Anthropic's format
            refined_output = []
            
            for action_item in parsed_output:
                
                print("Action Item:", action_item)
                # sometime showui returns lower case action names
                action_item["action"] = action_item["action"].upper()
                
                if action_item["action"] not in self.supported_action_type:
                    raise ValueError(f"Action {action_item['action']} not supported. Check the output from ShowUI: {output_text}")
                    # continue
                
                elif action_item["action"] == "CLICK":  # 1. click -> mouse_move + left_click
                    x, y = action_item["position"]
                    action_item["position"] = (int(x * (self.screen_bbox[2] - self.screen_bbox[0])),
                                               int(y * (self.screen_bbox[3] - self.screen_bbox[1])))
                    refined_output.append({"action": "mouse_move", "text": None, "coordinate": tuple(action_item["position"])})
                    refined_output.append({"action": "left_click", "text": None, "coordinate": None})
                
                elif action_item["action"] == "INPUT":  # 2. input -> type
                    refined_output.append({"action": "type", "text": action_item["value"], "coordinate": None})
                
                elif action_item["action"] == "ENTER":  # 3. enter -> key, enter
                    refined_output.append({"action": "key", "text": "Enter", "coordinate": None})
                
                elif action_item["action"] == "ESC" or action_item["action"] == "ESCAPE":  # 4. enter -> key, enter
                    refined_output.append({"action": "key", "text": "Escape", "coordinate": None})
                    
                elif action_item["action"] == "HOVER":  # 5. hover -> mouse_move
                    x, y = action_item["position"]
                    action_item["position"] = (int(x * (self.screen_bbox[2] - self.screen_bbox[0])),
                                               int(y * (self.screen_bbox[3] - self.screen_bbox[1])))
                    refined_output.append({"action": "mouse_move", "text": None, "coordinate": tuple(action_item["position"])})
                    
                elif action_item["action"] == "SCROLL":  # 6. scroll -> key: pagedown
                    if action_item["value"] == "up":
                        refined_output.append({"action": "key", "text": "pageup", "coordinate": None})
                    elif action_item["value"] == "down":
                        refined_output.append({"action": "key", "text": "pagedown", "coordinate": None})
                    else:
                        raise ValueError(f"Scroll direction {action_item['value']} not supported.")

                elif action_item["action"] == "PRESS":  # 7. press
                    x, y = action_item["position"]
                    action_item["position"] = (int(x * (self.screen_bbox[2] - self.screen_bbox[0])),
                                               int(y * (self.screen_bbox[3] - self.screen_bbox[1])))
                    refined_output.append({"action": "mouse_move", "text": None, "coordinate": tuple(action_item["position"])})
                    refined_output.append({"action": "left_press", "text": None, "coordinate": None})

            return refined_output

        except Exception as e:
            print(f"Error parsing output: {e}")
            return None
        
    def _get_screen_resolution(self):
        from screeninfo import get_monitors
        import platform
        if platform.system() == "Darwin":
            import Quartz  # uncomment this line if you are on macOS
        import subprocess
            
        # Detect platform
        system = platform.system()

        if system == "Windows":
            # Windows: Use screeninfo to get monitor details
            screens = get_monitors()

            # Sort screens by x position to arrange from left to right
            sorted_screens = sorted(screens, key=lambda s: s.x)

            if self.selected_screen < 0 or self.selected_screen >= len(screens):
                raise IndexError("Invalid screen index.")

            screen = sorted_screens[self.selected_screen]
            bbox = (screen.x, screen.y, screen.x + screen.width, screen.y + screen.height)

        elif system == "Darwin":  # macOS
            # macOS: Use Quartz to get monitor details
            max_displays = 32  # Maximum number of displays to handle
            active_displays = Quartz.CGGetActiveDisplayList(max_displays, None, None)[1]

            # Get the display bounds (resolution) for each active display
            screens = []
            for display_id in active_displays:
                bounds = Quartz.CGDisplayBounds(display_id)
                screens.append({
                    'id': display_id,
                    'x': int(bounds.origin.x),
                    'y': int(bounds.origin.y),
                    'width': int(bounds.size.width),
                    'height': int(bounds.size.height),
                    'is_primary': Quartz.CGDisplayIsMain(display_id)  # Check if this is the primary display
                })

            # Sort screens by x position to arrange from left to right
            sorted_screens = sorted(screens, key=lambda s: s['x'])

            if self.selected_screen < 0 or self.selected_screen >= len(screens):
                raise IndexError("Invalid screen index.")

            screen = sorted_screens[self.selected_screen]
            bbox = (screen['x'], screen['y'], screen['x'] + screen['width'], screen['y'] + screen['height'])

        else:  # Linux or other OS
            cmd = "xrandr | grep ' primary' | awk '{print $4}'"
            try:
                output = subprocess.check_output(cmd, shell=True).decode()
                resolution = output.strip().split()[0]
                width, height = map(int, resolution.split('x'))
                bbox = (0, 0, width, height)  # Assuming single primary screen for simplicity
            except subprocess.CalledProcessError:
                raise RuntimeError("Failed to get screen resolution on Linux.")
        
        return bbox



def _message_display_callback(messages):
    display_messages = []
    for msg in messages:
        try:
            if isinstance(msg["content"][0], TextBlock):
                display_messages.append((msg["content"][0].text, None))  # User message
            elif isinstance(msg["content"][0], BetaTextBlock):
                display_messages.append((None, msg["content"][0].text))  # Bot message
            elif isinstance(msg["content"][0], BetaToolUseBlock):
                display_messages.append((None, f"Tool Use: {msg['content'][0].name}\nInput: {msg['content'][0].input}"))  # Bot message
            elif isinstance(msg["content"][0], Dict) and msg["content"][0]["content"][-1]["type"] == "image":
                display_messages.append((None, f'<img src="data:image/png;base64,{msg["content"][0]["content"][-1]["source"]["data"]}">'))  # Bot message
            else:
                pass
                # print(msg["content"][0])
        except Exception as e:
            print("error", e)
            pass
    return display_messages


def _make_api_tool_result(
    result: ToolResult, tool_use_id: str
) -> BetaToolResultBlockParam:
    """Convert an agent ToolResult to an API ToolResultBlockParam."""
    tool_result_content: list[BetaTextBlockParam | BetaImageBlockParam] | str = []
    is_error = False
    if result.error:
        is_error = True
        tool_result_content = _maybe_prepend_system_tool_result(result, result.error)
    else:
        if result.output:
            tool_result_content.append(
                {
                    "type": "text",
                    "text": _maybe_prepend_system_tool_result(result, result.output),
                }
            )
        if result.base64_image:
            tool_result_content.append(
                {
                    "type": "image",
                    "source": {
                        "type": "base64",
                        "media_type": "image/png",
                        "data": result.base64_image,
                    },
                }
            )
    return {
        "type": "tool_result",
        "content": tool_result_content,
        "tool_use_id": tool_use_id,
        "is_error": is_error,
    }


def _maybe_prepend_system_tool_result(result: ToolResult, result_text: str):
    if result.system:
        result_text = f"<system>{result.system}</system>\n{result_text}"
    return result_text



# Testing main function
if __name__ == "__main__":
    def output_callback(content_block):
        # print("Output Callback:", content_block)
        pass

    def tool_output_callback(result, action):
        print("[showui_executor] Tool Output Callback:", result, action)
        pass

    # Instantiate the executor
    executor = ShowUIExecutor(
        output_callback=output_callback,
        tool_output_callback=tool_output_callback,
        selected_screen=0
    )

    # test inputs
    response_content = "{'content': \"{'action': 'CLICK', 'value': None, 'position': [0.49, 0.18]}\", 'role': 'assistant'}"
    # response_content = {'content': "{'action': 'CLICK', 'value': None, 'position': [0.49, 0.39]}", 'role': 'assistant'}
    # response_content = "{'content': \"{'action': 'CLICK', 'value': None, 'position': [0.49, 0.42]}, {'action': 'INPUT', 'value': 'weather for New York city', 'position': [0.49, 0.42]}, {'action': 'ENTER', 'value': None, 'position': None}\", 'role': 'assistant'}"

    # Initialize messages
    messages = []

    # Call the executor
    print("Testing ShowUIExecutor with response content:", response_content)
    for message, tool_result_content in executor(response_content, messages):
        print("Message:", message)
        print("Tool Result Content:", tool_result_content)

    # Display final messages
    print("\nFinal messages:")
    for msg in messages:
        print(msg)
        
        

[
    {'role': 'user', 'content': ['open a new tab and go to amazon.com', 'tmp/outputs/screenshot_b4a1b7e60a5c47359bedbd8707573966.png']},
    {'role': 'assistant', 'content': ["History Action: {'action': 'mouse_move', 'text': None, 'coordinate': (1216, 88)}"]}, 
    {'role': 'assistant', 'content': ["History Action: {'action': 'left_click', 'text': None, 'coordinate': None}"]},
    {'content': [
        {'type': 'tool_result', 'content': [{'type': 'text', 'text': 'Moved mouse to (1216, 88)'}], 'tool_use_id': 'toolu_ae4f2886-366c-4789-9fa6-ec13461cef12', 'is_error': False},
        {'type': 'tool_result', 'content': [{'type': 'text', 'text': 'Performed left_click'}], 'tool_use_id': 'toolu_a7377954-e1b7-4746-9757-b2eb4dcddc82', 'is_error': False}
                ], 'role': 'user'}
]